/* Generated by Cython 0.19-dev on Wed Feb 27 12:45:00 2013 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02040000
    #error Cython requires Python 2.4+.
#else
#include <stddef.h> /* For offsetof */
#ifndef offsetof
#define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
#define CYTHON_COMPILING_IN_PYPY 1
#define CYTHON_COMPILING_IN_CPYTHON 0
#else
#define CYTHON_COMPILING_IN_PYPY 0
#define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if PY_VERSION_HEX < 0x02050000
  typedef int Py_ssize_t;
  #define PY_SSIZE_T_MAX INT_MAX
  #define PY_SSIZE_T_MIN INT_MIN
  #define PY_FORMAT_SIZE_T ""
  #define CYTHON_FORMAT_SSIZE_T ""
  #define PyInt_FromSsize_t(z) PyInt_FromLong(z)
  #define PyInt_AsSsize_t(o)   __Pyx_PyInt_AsInt(o)
  #define PyNumber_Index(o)    ((PyNumber_Check(o) && !PyFloat_Check(o)) ? PyNumber_Int(o) : \
                                (PyErr_Format(PyExc_TypeError, \
                                              "expected index value, got %.200s", Py_TYPE(o)->tp_name), \
                                 (PyObject*)0))
  #define __Pyx_PyIndex_Check(o) (PyNumber_Check(o) && !PyFloat_Check(o) && \
                                  !PyComplex_Check(o))
  #define PyIndex_Check __Pyx_PyIndex_Check
  #define PyErr_WarnEx(category, message, stacklevel) PyErr_Warn(category, message)
  #define __PYX_BUILD_PY_SSIZE_T "i"
#else
  #define __PYX_BUILD_PY_SSIZE_T "n"
  #define CYTHON_FORMAT_SSIZE_T "z"
  #define __Pyx_PyIndex_Check PyIndex_Check
#endif
#if PY_VERSION_HEX < 0x02060000
  #define Py_REFCNT(ob) (((PyObject*)(ob))->ob_refcnt)
  #define Py_TYPE(ob)   (((PyObject*)(ob))->ob_type)
  #define Py_SIZE(ob)   (((PyVarObject*)(ob))->ob_size)
  #define PyVarObject_HEAD_INIT(type, size) \
          PyObject_HEAD_INIT(type) size,
  #define PyType_Modified(t)
  typedef struct {
     void *buf;
     PyObject *obj;
     Py_ssize_t len;
     Py_ssize_t itemsize;
     int readonly;
     int ndim;
     char *format;
     Py_ssize_t *shape;
     Py_ssize_t *strides;
     Py_ssize_t *suboffsets;
     void *internal;
  } Py_buffer;
  #define PyBUF_SIMPLE 0
  #define PyBUF_WRITABLE 0x0001
  #define PyBUF_FORMAT 0x0004
  #define PyBUF_ND 0x0008
  #define PyBUF_STRIDES (0x0010 | PyBUF_ND)
  #define PyBUF_C_CONTIGUOUS (0x0020 | PyBUF_STRIDES)
  #define PyBUF_F_CONTIGUOUS (0x0040 | PyBUF_STRIDES)
  #define PyBUF_ANY_CONTIGUOUS (0x0080 | PyBUF_STRIDES)
  #define PyBUF_INDIRECT (0x0100 | PyBUF_STRIDES)
  #define PyBUF_RECORDS (PyBUF_STRIDES | PyBUF_FORMAT | PyBUF_WRITABLE)
  #define PyBUF_FULL (PyBUF_INDIRECT | PyBUF_FORMAT | PyBUF_WRITABLE)
  typedef int (*getbufferproc)(PyObject *, Py_buffer *, int);
  typedef void (*releasebufferproc)(PyObject *, Py_buffer *);
#endif
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#if PY_MAJOR_VERSION < 3 && PY_MINOR_VERSION < 6
  #define PyUnicode_FromString(s) PyUnicode_Decode(s, strlen(s), "UTF-8", "strict")
#endif
#if PY_MAJOR_VERSION >= 3
  #define Py_TPFLAGS_CHECKTYPES 0
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#if (PY_VERSION_HEX < 0x02060000) || (PY_MAJOR_VERSION >= 3)
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ? \
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((k=k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_VERSION_HEX < 0x02060000
  #define PyBytesObject                PyStringObject
  #define PyBytes_Type                 PyString_Type
  #define PyBytes_Check                PyString_Check
  #define PyBytes_CheckExact           PyString_CheckExact
  #define PyBytes_FromString           PyString_FromString
  #define PyBytes_FromStringAndSize    PyString_FromStringAndSize
  #define PyBytes_FromFormat           PyString_FromFormat
  #define PyBytes_DecodeEscape         PyString_DecodeEscape
  #define PyBytes_AsString             PyString_AsString
  #define PyBytes_AsStringAndSize      PyString_AsStringAndSize
  #define PyBytes_Size                 PyString_Size
  #define PyBytes_AS_STRING            PyString_AS_STRING
  #define PyBytes_GET_SIZE             PyString_GET_SIZE
  #define PyBytes_Repr                 PyString_Repr
  #define PyBytes_Concat               PyString_Concat
  #define PyBytes_ConcatAndDel         PyString_ConcatAndDel
#endif
#if PY_VERSION_HEX < 0x02060000
  #define PySet_Check(obj)             PyObject_TypeCheck(obj, &PySet_Type)
  #define PyFrozenSet_Check(obj)       PyObject_TypeCheck(obj, &PyFrozenSet_Type)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_VERSION_HEX < 0x03020000
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if (PY_MAJOR_VERSION < 3) || (PY_VERSION_HEX >= 0x03010300)
  #define __Pyx_PySequence_GetSlice(obj, a, b) PySequence_GetSlice(obj, a, b)
  #define __Pyx_PySequence_SetSlice(obj, a, b, value) PySequence_SetSlice(obj, a, b, value)
  #define __Pyx_PySequence_DelSlice(obj, a, b) PySequence_DelSlice(obj, a, b)
#else
  #define __Pyx_PySequence_GetSlice(obj, a, b) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), (PyObject*)0) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_GetSlice(obj, a, b)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object is unsliceable", (obj)->ob_type->tp_name), (PyObject*)0)))
  #define __Pyx_PySequence_SetSlice(obj, a, b, value) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), -1) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_SetSlice(obj, a, b, value)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object doesn't support slice assignment", (obj)->ob_type->tp_name), -1)))
  #define __Pyx_PySequence_DelSlice(obj, a, b) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), -1) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_DelSlice(obj, a, b)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object doesn't support slice deletion", (obj)->ob_type->tp_name), -1)))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#endif
#if PY_VERSION_HEX < 0x02050000
  #define __Pyx_GetAttrString(o,n)   PyObject_GetAttrString((o),((char *)(n)))
  #define __Pyx_SetAttrString(o,n,a) PyObject_SetAttrString((o),((char *)(n)),(a))
  #define __Pyx_DelAttrString(o,n)   PyObject_DelAttrString((o),((char *)(n)))
#else
  #define __Pyx_GetAttrString(o,n)   PyObject_GetAttrString((o),(n))
  #define __Pyx_SetAttrString(o,n,a) PyObject_SetAttrString((o),(n),(a))
  #define __Pyx_DelAttrString(o,n)   PyObject_DelAttrString((o),(n))
#endif
#if PY_VERSION_HEX < 0x02050000
  #define __Pyx_NAMESTR(n) ((char *)(n))
  #define __Pyx_DOCSTR(n)  ((char *)(n))
#else
  #define __Pyx_NAMESTR(n) (n)
  #define __Pyx_DOCSTR(n)  (n)
#endif


#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
#define _USE_MATH_DEFINES
#endif
#include <math.h>
#define __PYX_HAVE__numba__extension_types
#define __PYX_HAVE_API__numba__extension_types
#include "string.h"
#include "stdio.h"
#include "pythread.h"
#include "numbafunction.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; char *s; const long n; const char* encoding; const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry; /*proto*/

#define __Pyx_PyBytes_FromUString(s) PyBytes_FromString((char*)s)
#define __Pyx_PyBytes_AsUString(s)   ((unsigned char*) PyBytes_AsString(s))
#define __Pyx_Owned_Py_None(b) (Py_INCREF(Py_None), Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? (Py_INCREF(Py_True), Py_True) : (Py_INCREF(Py_False), Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE size_t __Pyx_PyInt_AsSize_t(PyObject*);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))


#ifdef __GNUC__
  /* Test for GCC > 2.95 */
  #if __GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95))
    #define likely(x)   __builtin_expect(!!(x), 1)
    #define unlikely(x) __builtin_expect(!!(x), 0)
  #else /* __GNUC__ > 2 ... */
    #define likely(x)   (x)
    #define unlikely(x) (x)
  #endif /* __GNUC__ > 2 ... */
#else /* __GNUC__ */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
    
static PyObject *__pyx_m;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "extension_types.pyx",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type;

/* "numba/extension_types.pyx":15
 * import numba
 * 
 * ctypedef object (*tp_new_func)(PyObject *, PyObject *, PyObject *)             # <<<<<<<<<<<<<<
 * ctypedef void (*destructor)(PyObject *)
 * ctypedef int (*visitproc)(PyObject *, void *)
 */
typedef PyObject *(*__pyx_t_5numba_15extension_types_tp_new_func)(PyObject *, PyObject *, PyObject *);

/* "numba/extension_types.pyx":16
 * 
 * ctypedef object (*tp_new_func)(PyObject *, PyObject *, PyObject *)
 * ctypedef void (*destructor)(PyObject *)             # <<<<<<<<<<<<<<
 * ctypedef int (*visitproc)(PyObject *, void *)
 * 
 */
typedef void (*__pyx_t_5numba_15extension_types_destructor)(PyObject *);

/* "numba/extension_types.pyx":17
 * ctypedef object (*tp_new_func)(PyObject *, PyObject *, PyObject *)
 * ctypedef void (*destructor)(PyObject *)
 * ctypedef int (*visitproc)(PyObject *, void *)             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
typedef int (*__pyx_t_5numba_15extension_types_visitproc)(PyObject *, void *);

/* "numba/extension_types.pyx":184
 * #------------------------------------------------------------------------
 * 
 * def create_new_extension_type(name, bases, dict, ext_numba_type,             # <<<<<<<<<<<<<<
 *                               vtab, vtab_type, llvm_methods, method_pointers):
 *     """
 */
struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type {
  PyObject_HEAD
  Py_ssize_t __pyx_v_attrs_offset;
  int __pyx_v_base_is_object;
  PyObject *__pyx_v_ext_type;
  PyObject *__pyx_v_orig_new;
  PyObject *__pyx_v_vtab;
  Py_ssize_t __pyx_v_vtab_offset;
};

#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname); /*proto*/
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          if (acquire_gil) { \
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure(); \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
              PyGILState_Release(__pyx_gilstate_save); \
          } else { \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext() \
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif /* CYTHON_REFNANNY */
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

static PyObject *__Pyx_GetName(PyObject *dict, PyObject *name); /*proto*/

static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *); /*proto*/

static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found); /*proto*/

static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name); /*proto*/

static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[], \
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args, \
    const char* function_name); /*proto*/

static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

static CYTHON_INLINE int __Pyx_IterFinish(void); /*proto*/

static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected); /*proto*/

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
#define __Pyx_GetItemInt_List(o, i, size, to_py_func) (((size) <= sizeof(Py_ssize_t)) ? \
                                                    __Pyx_GetItemInt_List_Fast(o, i) : \
                                                    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    else if ((-PyList_GET_SIZE(o) <= i) & (i < 0)) {
        PyObject *r = PyList_GET_ITEM(o, PyList_GET_SIZE(o) + i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
#define __Pyx_GetItemInt_Tuple(o, i, size, to_py_func) (((size) <= sizeof(Py_ssize_t)) ? \
                                                    __Pyx_GetItemInt_Tuple_Fast(o, i) : \
                                                    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    else if ((-PyTuple_GET_SIZE(o) <= i) & (i < 0)) {
        PyObject *r = PyTuple_GET_ITEM(o, PyTuple_GET_SIZE(o) + i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
#define __Pyx_GetItemInt(o, i, size, to_py_func) (((size) <= sizeof(Py_ssize_t)) ? \
                                                    __Pyx_GetItemInt_Fast(o, i) : \
                                                    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (PyList_CheckExact(o)) {
        Py_ssize_t n = (likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if (likely((n >= 0) & (n < PyList_GET_SIZE(o)))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = (likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if (likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {  /* inlined PySequence_GetItem() */
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (unlikely(l < 0)) return NULL;
                i += l;
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f) \
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f) \
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f) \
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g) \
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
    int flags;
    PyObject *func_dict;
    PyObject *func_weakreflist;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj; /* No-args super() class cell */
    void *defaults;
    int defaults_pyobjects;
    PyObject *defaults_tuple; /* Const defaults tuple */
    PyObject *(*defaults_getter)(PyObject *);
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, code) \
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self, PyObject *module,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static int __Pyx_CyFunction_init(void);

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb); /*proto*/
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb); /*proto*/

static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause); /*proto*/

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level); /*proto*/

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_Py_uintptr_t(Py_uintptr_t);

static CYTHON_INLINE Py_uintptr_t __Pyx_PyInt_from_py_Py_uintptr_t(PyObject *);

static CYTHON_INLINE unsigned char __Pyx_PyInt_AsUnsignedChar(PyObject *);

static CYTHON_INLINE unsigned short __Pyx_PyInt_AsUnsignedShort(PyObject *);

static CYTHON_INLINE unsigned int __Pyx_PyInt_AsUnsignedInt(PyObject *);

static CYTHON_INLINE char __Pyx_PyInt_AsChar(PyObject *);

static CYTHON_INLINE short __Pyx_PyInt_AsShort(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_AsInt(PyObject *);

static CYTHON_INLINE signed char __Pyx_PyInt_AsSignedChar(PyObject *);

static CYTHON_INLINE signed short __Pyx_PyInt_AsSignedShort(PyObject *);

static CYTHON_INLINE signed int __Pyx_PyInt_AsSignedInt(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_AsLongDouble(PyObject *);

static CYTHON_INLINE unsigned long __Pyx_PyInt_AsUnsignedLong(PyObject *);

static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_AsUnsignedLongLong(PyObject *);

static CYTHON_INLINE long __Pyx_PyInt_AsLong(PyObject *);

static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_AsLongLong(PyObject *);

static CYTHON_INLINE signed long __Pyx_PyInt_AsSignedLong(PyObject *);

static CYTHON_INLINE signed PY_LONG_LONG __Pyx_PyInt_AsSignedLongLong(PyObject *);

static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename); /*proto*/

static int __Pyx_check_binary_version(void);

#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

static PyObject *__Pyx_ImportModule(const char *name); /*proto*/

static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);  /*proto*/

typedef struct {
    int code_line;
    PyCodeObject* code_object;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename); /*proto*/

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t); /*proto*/


/* Module declarations from 'cython' */

/* Module declarations from 'cpython.version' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'numba._numba' */

/* Module declarations from 'numba.extension_types' */
static PyTypeObject *__pyx_ptype_5numba_15extension_types___pyx_scope_struct__create_new_extension_type = 0;
static Py_uintptr_t __pyx_f_5numba_15extension_types_align(Py_uintptr_t, size_t); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numba_15extension_types_getoffsets(PyObject *); /*proto*/
static CYTHON_INLINE void __pyx_f_5numba_15extension_types_setnone(PyObject **); /*proto*/
#define __Pyx_MODULE_NAME "numba.extension_types"
int __pyx_module_is_main_numba__extension_types = 0;

/* Implementation of 'numba.extension_types' */
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_pf_5numba_15extension_types_compute_vtab_offset(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_py_class); /* proto */
static PyObject *__pyx_pf_5numba_15extension_types_2compute_attrs_offset(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_py_class); /* proto */
static PyObject *__pyx_pf_5numba_15extension_types_4compute_object_offsets(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ext_type, PyObject *__pyx_v_struct_offset); /* proto */
static PyObject *__pyx_pf_5numba_15extension_types_25create_new_extension_type_new(PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_args, PyObject *__pyx_v_kwds); /* proto */
static PyObject *__pyx_pf_5numba_15extension_types_6create_new_extension_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_bases, PyObject *__pyx_v_dict, PyObject *__pyx_v_ext_numba_type, PyObject *__pyx_v_vtab, PyObject *__pyx_v_vtab_type, PyObject *__pyx_v_llvm_methods, PyObject *__pyx_v_method_pointers); /* proto */
static PyObject *__pyx_pf_5numba_15extension_types_8create_function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_methoddef, PyObject *__pyx_v_py_func, PyObject *__pyx_v_lfunc_pointer, PyObject *__pyx_v_signature, PyObject *__pyx_v_modname); /* proto */
static PyObject *__pyx_tp_new_5numba_15extension_types___pyx_scope_struct__create_new_extension_type(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static char __pyx_k_1[] = "__numba_object_offset";
static char __pyx_k_2[] = "__numba_struct_ctype_p";
static char __pyx_k_6[] = "/Users/mark/numba/numba/extension_types.pyx";
static char __pyx_k_7[] = "create_new_extension_type.<locals>.new";
static char __pyx_k_8[] = "numba.extension_types";
static char __pyx_k_9[] = "compute_attrs_offset";
static char __pyx_k_10[] = "Subclassing variable sized objects";
static char __pyx_k_12[] = "__numba_method_pointers";
static char __pyx_k_13[] = "compute_object_offsets";
static char __pyx_k_14[] = "\nThis modules creates extension types.\n\nSee also numba.extension_type_inference\n";
static char __pyx_k_15[] = "NumbaFunction_NewEx_pointer";
static char __pyx_k_16[] = "numbafunc_closure_field_offset";
static char __pyx_k_25[] = "create_new_extension_type";
static char __pyx_k__ml[] = "ml";
static char __pyx_k__cls[] = "cls";
static char __pyx_k__get[] = "get";
static char __pyx_k__new[] = "new";
static char __pyx_k__obj[] = "obj";
static char __pyx_k__sys[] = "sys";
static char __pyx_k__args[] = "args";
static char __pyx_k__cast[] = "cast";
static char __pyx_k__dict[] = "dict";
static char __pyx_k__kwds[] = "kwds";
static char __pyx_k__name[] = "name";
static char __pyx_k__vtab[] = "vtab";
static char __pyx_k__bases[] = "bases";
static char __pyx_k__byref[] = "byref";
static char __pyx_k__numba[] = "numba";
static char __pyx_k__obj_p[] = "obj_p";
static char __pyx_k__super[] = "super";
static char __pyx_k__upper[] = "upper";
static char __pyx_k__value[] = "value";
static char __pyx_k__ctypes[] = "ctypes";
static char __pyx_k__fields[] = "fields";
static char __pyx_k__object[] = "object";
static char __pyx_k__offset[] = "offset";
static char __pyx_k__result[] = "result";
static char __pyx_k__sizeof[] = "sizeof";
static char __pyx_k__type_p[] = "type_p";
static char __pyx_k__vtab_p[] = "vtab_p";
static char __pyx_k____new__[] = "__new__";
static char __pyx_k__exttype[] = "exttype";
static char __pyx_k__lfunc_p[] = "lfunc_p";
static char __pyx_k__modname[] = "modname";
static char __pyx_k__offsets[] = "offsets";
static char __pyx_k__pointer[] = "pointer";
static char __pyx_k__py_func[] = "py_func";
static char __pyx_k____main__[] = "__main__";
static char __pyx_k____test__[] = "__test__";
static char __pyx_k__c_void_p[] = "c_void_p";
static char __pyx_k__ext_type[] = "ext_type";
static char __pyx_k__is_array[] = "is_array";
static char __pyx_k__new_func[] = "new_func";
static char __pyx_k__offsetof[] = "offsetof";
static char __pyx_k__orig_new[] = "orig_new";
static char __pyx_k__py_class[] = "py_class";
static char __pyx_k__func_code[] = "func_code";
static char __pyx_k__is_object[] = "is_object";
static char __pyx_k__methoddef[] = "methoddef";
static char __pyx_k__signature[] = "signature";
static char __pyx_k__to_ctypes[] = "to_ctypes";
static char __pyx_k__vtab_type[] = "vtab_type";
static char __pyx_k__ext_type_p[] = "ext_type_p";
static char __pyx_k__field_name[] = "field_name";
static char __pyx_k__field_type[] = "field_type";
static char __pyx_k__methoddef_p[] = "methoddef_p";
static char __pyx_k__struct_type[] = "struct_type";
static char __pyx_k__vtab_offset[] = "vtab_offset";
static char __pyx_k____numba_vtab[] = "__numba_vtab";
static char __pyx_k___numba_attrs[] = "_numba_attrs";
static char __pyx_k__attrs_offset[] = "attrs_offset";
static char __pyx_k__llvm_methods[] = "llvm_methods";
static char __pyx_k__staticmethod[] = "staticmethod";
static char __pyx_k__struct_ctype[] = "struct_ctype";
static char __pyx_k__attrs_pointer[] = "attrs_pointer";
static char __pyx_k__lfunc_pointer[] = "lfunc_pointer";
static char __pyx_k__struct_offset[] = "struct_offset";
static char __pyx_k__vtab_location[] = "vtab_location";
static char __pyx_k____numba_lfuncs[] = "__numba_lfuncs";
static char __pyx_k____numba_vtab_p[] = "__numba_vtab_p";
static char __pyx_k__base_is_object[] = "base_is_object";
static char __pyx_k__ext_numba_type[] = "ext_numba_type";
static char __pyx_k____numba_obj_end[] = "__numba_obj_end";
static char __pyx_k__create_function[] = "create_function";
static char __pyx_k__method_pointers[] = "method_pointers";
static char __pyx_k____numba_ext_type[] = "__numba_ext_type";
static char __pyx_k__attribute_struct[] = "attribute_struct";
static char __pyx_k____numba_vtab_type[] = "__numba_vtab_type";
static char __pyx_k__NotImplementedError[] = "NotImplementedError";
static char __pyx_k____numba_attr_offset[] = "__numba_attr_offset";
static char __pyx_k____numba_orig_tp_new[] = "__numba_orig_tp_new";
static char __pyx_k____numba_struct_type[] = "__numba_struct_type";
static char __pyx_k____numba_vtab_offset[] = "__numba_vtab_offset";
static char __pyx_k__compute_vtab_offset[] = "compute_vtab_offset";
static PyObject *__pyx_n_s_1;
static PyObject *__pyx_kp_s_10;
static PyObject *__pyx_n_s_12;
static PyObject *__pyx_n_s_13;
static PyObject *__pyx_n_s_15;
static PyObject *__pyx_n_s_16;
static PyObject *__pyx_n_s_2;
static PyObject *__pyx_n_s_25;
static PyObject *__pyx_kp_s_6;
static PyObject *__pyx_n_s_7;
static PyObject *__pyx_n_s_8;
static PyObject *__pyx_n_s_9;
static PyObject *__pyx_n_s__NotImplementedError;
static PyObject *__pyx_n_s____main__;
static PyObject *__pyx_n_s____new__;
static PyObject *__pyx_n_s____numba_attr_offset;
static PyObject *__pyx_n_s____numba_ext_type;
static PyObject *__pyx_n_s____numba_lfuncs;
static PyObject *__pyx_n_s____numba_obj_end;
static PyObject *__pyx_n_s____numba_orig_tp_new;
static PyObject *__pyx_n_s____numba_struct_type;
static PyObject *__pyx_n_s____numba_vtab;
static PyObject *__pyx_n_s____numba_vtab_offset;
static PyObject *__pyx_n_s____numba_vtab_p;
static PyObject *__pyx_n_s____numba_vtab_type;
static PyObject *__pyx_n_s____test__;
static PyObject *__pyx_n_s___numba_attrs;
static PyObject *__pyx_n_s__args;
static PyObject *__pyx_n_s__attribute_struct;
static PyObject *__pyx_n_s__attrs_offset;
static PyObject *__pyx_n_s__attrs_pointer;
static PyObject *__pyx_n_s__base_is_object;
static PyObject *__pyx_n_s__bases;
static PyObject *__pyx_n_s__byref;
static PyObject *__pyx_n_s__c_void_p;
static PyObject *__pyx_n_s__cast;
static PyObject *__pyx_n_s__cls;
static PyObject *__pyx_n_s__compute_vtab_offset;
static PyObject *__pyx_n_s__create_function;
static PyObject *__pyx_n_s__ctypes;
static PyObject *__pyx_n_s__dict;
static PyObject *__pyx_n_s__ext_numba_type;
static PyObject *__pyx_n_s__ext_type;
static PyObject *__pyx_n_s__ext_type_p;
static PyObject *__pyx_n_s__exttype;
static PyObject *__pyx_n_s__field_name;
static PyObject *__pyx_n_s__field_type;
static PyObject *__pyx_n_s__fields;
static PyObject *__pyx_n_s__func_code;
static PyObject *__pyx_n_s__get;
static PyObject *__pyx_n_s__is_array;
static PyObject *__pyx_n_s__is_object;
static PyObject *__pyx_n_s__kwds;
static PyObject *__pyx_n_s__lfunc_p;
static PyObject *__pyx_n_s__lfunc_pointer;
static PyObject *__pyx_n_s__llvm_methods;
static PyObject *__pyx_n_s__method_pointers;
static PyObject *__pyx_n_s__methoddef;
static PyObject *__pyx_n_s__methoddef_p;
static PyObject *__pyx_n_s__ml;
static PyObject *__pyx_n_s__modname;
static PyObject *__pyx_n_s__name;
static PyObject *__pyx_n_s__new;
static PyObject *__pyx_n_s__new_func;
static PyObject *__pyx_n_s__numba;
static PyObject *__pyx_n_s__obj;
static PyObject *__pyx_n_s__obj_p;
static PyObject *__pyx_n_s__object;
static PyObject *__pyx_n_s__offset;
static PyObject *__pyx_n_s__offsetof;
static PyObject *__pyx_n_s__offsets;
static PyObject *__pyx_n_s__orig_new;
static PyObject *__pyx_n_s__pointer;
static PyObject *__pyx_n_s__py_class;
static PyObject *__pyx_n_s__py_func;
static PyObject *__pyx_n_s__result;
static PyObject *__pyx_n_s__signature;
static PyObject *__pyx_n_s__sizeof;
static PyObject *__pyx_n_s__staticmethod;
static PyObject *__pyx_n_s__struct_ctype;
static PyObject *__pyx_n_s__struct_offset;
static PyObject *__pyx_n_s__struct_type;
static PyObject *__pyx_n_s__super;
static PyObject *__pyx_n_s__sys;
static PyObject *__pyx_n_s__to_ctypes;
static PyObject *__pyx_n_s__type_p;
static PyObject *__pyx_n_s__upper;
static PyObject *__pyx_n_s__value;
static PyObject *__pyx_n_s__vtab;
static PyObject *__pyx_n_s__vtab_location;
static PyObject *__pyx_n_s__vtab_offset;
static PyObject *__pyx_n_s__vtab_p;
static PyObject *__pyx_n_s__vtab_type;
static PyObject *__pyx_k_tuple_3;
static PyObject *__pyx_k_tuple_4;
static PyObject *__pyx_k_tuple_11;
static PyObject *__pyx_k_tuple_17;
static PyObject *__pyx_k_tuple_19;
static PyObject *__pyx_k_tuple_21;
static PyObject *__pyx_k_tuple_23;
static PyObject *__pyx_k_tuple_26;
static PyObject *__pyx_k_codeobj_5;
static PyObject *__pyx_k_codeobj_18;
static PyObject *__pyx_k_codeobj_20;
static PyObject *__pyx_k_codeobj_22;
static PyObject *__pyx_k_codeobj_24;
static PyObject *__pyx_k_codeobj_27;

/* "numba/extension_types.pyx":46
 * numbafunc_closure_field_offset = closure_field_offset
 * 
 * cdef Py_uintptr_t align(Py_uintptr_t p, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align on a boundary"
 *     cdef size_t offset
 */

static Py_uintptr_t __pyx_f_5numba_15extension_types_align(Py_uintptr_t __pyx_v_p, size_t __pyx_v_alignment) {
  size_t __pyx_v_offset;
  Py_uintptr_t __pyx_r;
  int __pyx_t_1;

  /* "numba/extension_types.pyx":51
 * 
 *     with cython.cdivision(True):
 *         offset = p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */
  __pyx_v_offset = (__pyx_v_p % __pyx_v_alignment);

  /* "numba/extension_types.pyx":53
 *         offset = p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         p += alignment - offset
 * 
 */
  __pyx_t_1 = (__pyx_v_offset > 0);
  if (__pyx_t_1) {

    /* "numba/extension_types.pyx":54
 * 
 *     if offset > 0:
 *         p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return p
 */
    __pyx_v_p = (__pyx_v_p + (__pyx_v_alignment - __pyx_v_offset));
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numba/extension_types.pyx":56
 *         p += alignment - offset
 * 
 *     return p             # <<<<<<<<<<<<<<
 * 
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "numba/extension_types.pyx":58
 *     return p
 * 
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     return <void *> align(<Py_uintptr_t> memory, alignment)
 */

static void *__pyx_f_5numba_15extension_types_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  void *__pyx_r;

  /* "numba/extension_types.pyx":60
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     return <void *> align(<Py_uintptr_t> memory, alignment)             # <<<<<<<<<<<<<<
 * 
 * def compute_vtab_offset(py_class):
 */
  __pyx_r = ((void *)__pyx_f_5numba_15extension_types_align(((Py_uintptr_t)__pyx_v_memory), __pyx_v_alignment));
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5numba_15extension_types_1compute_vtab_offset(PyObject *__pyx_self, PyObject *__pyx_v_py_class); /*proto*/
static char __pyx_doc_5numba_15extension_types_compute_vtab_offset[] = "Returns the vtab pointer offset in the object";
static PyMethodDef __pyx_mdef_5numba_15extension_types_1compute_vtab_offset = {__Pyx_NAMESTR("compute_vtab_offset"), (PyCFunction)__pyx_pw_5numba_15extension_types_1compute_vtab_offset, METH_O, __Pyx_DOCSTR(__pyx_doc_5numba_15extension_types_compute_vtab_offset)};
static PyObject *__pyx_pw_5numba_15extension_types_1compute_vtab_offset(PyObject *__pyx_self, PyObject *__pyx_v_py_class) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_vtab_offset (wrapper)", 0);
  __pyx_r = __pyx_pf_5numba_15extension_types_compute_vtab_offset(__pyx_self, ((PyObject *)__pyx_v_py_class));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numba/extension_types.pyx":62
 *     return <void *> align(<Py_uintptr_t> memory, alignment)
 * 
 * def compute_vtab_offset(py_class):             # <<<<<<<<<<<<<<
 *     "Returns the vtab pointer offset in the object"
 *     offset = getattr(py_class, '__numba_vtab_offset', None)
 */

static PyObject *__pyx_pf_5numba_15extension_types_compute_vtab_offset(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_py_class) {
  PyObject *__pyx_v_offset = NULL;
  PyTypeObject *__pyx_v_type_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_vtab_offset", 0);

  /* "numba/extension_types.pyx":64
 * def compute_vtab_offset(py_class):
 *     "Returns the vtab pointer offset in the object"
 *     offset = getattr(py_class, '__numba_vtab_offset', None)             # <<<<<<<<<<<<<<
 *     if offset:
 *         return offset
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_py_class, ((PyObject *)__pyx_n_s____numba_vtab_offset), Py_None); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 64; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_offset = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":65
 *     "Returns the vtab pointer offset in the object"
 *     offset = getattr(py_class, '__numba_vtab_offset', None)
 *     if offset:             # <<<<<<<<<<<<<<
 *         return offset
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_offset); if (unlikely(__pyx_t_2 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 65; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_2) {

    /* "numba/extension_types.pyx":66
 *     offset = getattr(py_class, '__numba_vtab_offset', None)
 *     if offset:
 *         return offset             # <<<<<<<<<<<<<<
 * 
 *     cdef PyTypeObject *type_p = <PyTypeObject *> py_class
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_offset);
    __pyx_r = __pyx_v_offset;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numba/extension_types.pyx":68
 *         return offset
 * 
 *     cdef PyTypeObject *type_p = <PyTypeObject *> py_class             # <<<<<<<<<<<<<<
 *     return align(type_p.tp_basicsize, 8)
 * 
 */
  __pyx_v_type_p = ((PyTypeObject *)__pyx_v_py_class);

  /* "numba/extension_types.pyx":69
 * 
 *     cdef PyTypeObject *type_p = <PyTypeObject *> py_class
 *     return align(type_p.tp_basicsize, 8)             # <<<<<<<<<<<<<<
 * 
 * def compute_attrs_offset(py_class):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_uintptr_t(__pyx_f_5numba_15extension_types_align(__pyx_v_type_p->tp_basicsize, 8)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 69; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numba.extension_types.compute_vtab_offset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5numba_15extension_types_3compute_attrs_offset(PyObject *__pyx_self, PyObject *__pyx_v_py_class); /*proto*/
static char __pyx_doc_5numba_15extension_types_2compute_attrs_offset[] = "Returns the start of the attribute struct";
static PyMethodDef __pyx_mdef_5numba_15extension_types_3compute_attrs_offset = {__Pyx_NAMESTR("compute_attrs_offset"), (PyCFunction)__pyx_pw_5numba_15extension_types_3compute_attrs_offset, METH_O, __Pyx_DOCSTR(__pyx_doc_5numba_15extension_types_2compute_attrs_offset)};
static PyObject *__pyx_pw_5numba_15extension_types_3compute_attrs_offset(PyObject *__pyx_self, PyObject *__pyx_v_py_class) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_attrs_offset (wrapper)", 0);
  __pyx_r = __pyx_pf_5numba_15extension_types_2compute_attrs_offset(__pyx_self, ((PyObject *)__pyx_v_py_class));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numba/extension_types.pyx":71
 *     return align(type_p.tp_basicsize, 8)
 * 
 * def compute_attrs_offset(py_class):             # <<<<<<<<<<<<<<
 *     "Returns the start of the attribute struct"
 *     offset = getattr(py_class, '__numba_attr_offset', None)
 */

static PyObject *__pyx_pf_5numba_15extension_types_2compute_attrs_offset(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_py_class) {
  PyObject *__pyx_v_offset = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_uintptr_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_attrs_offset", 0);

  /* "numba/extension_types.pyx":73
 * def compute_attrs_offset(py_class):
 *     "Returns the start of the attribute struct"
 *     offset = getattr(py_class, '__numba_attr_offset', None)             # <<<<<<<<<<<<<<
 *     if offset:
 *         return offset
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_py_class, ((PyObject *)__pyx_n_s____numba_attr_offset), Py_None); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 73; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_offset = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":74
 *     "Returns the start of the attribute struct"
 *     offset = getattr(py_class, '__numba_attr_offset', None)
 *     if offset:             # <<<<<<<<<<<<<<
 *         return offset
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_offset); if (unlikely(__pyx_t_2 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 74; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_2) {

    /* "numba/extension_types.pyx":75
 *     offset = getattr(py_class, '__numba_attr_offset', None)
 *     if offset:
 *         return offset             # <<<<<<<<<<<<<<
 * 
 *     return align(compute_vtab_offset(py_class) + sizeof(void *), 8)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_offset);
    __pyx_r = __pyx_v_offset;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numba/extension_types.pyx":77
 *         return offset
 * 
 *     return align(compute_vtab_offset(py_class) + sizeof(void *), 8)             # <<<<<<<<<<<<<<
 * 
 * #------------------------------------------------------------------------
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__compute_vtab_offset); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_py_class);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_py_class);
  __Pyx_GIVEREF(__pyx_v_py_class);
  __pyx_t_4 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_FromSize_t((sizeof(void *))); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyNumber_Add(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_from_py_Py_uintptr_t(__pyx_t_1); if (unlikely((__pyx_t_5 == (Py_uintptr_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_uintptr_t(__pyx_f_5numba_15extension_types_align(__pyx_t_5, 8)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numba.extension_types.compute_attrs_offset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5numba_15extension_types_5compute_object_offsets(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5numba_15extension_types_4compute_object_offsets[] = "\n    Given an typesystem.ExtensionType, compute the offsets of object\n    pointers, relative to the start of the attribute offsets.\n    ";
static PyMethodDef __pyx_mdef_5numba_15extension_types_5compute_object_offsets = {__Pyx_NAMESTR("compute_object_offsets"), (PyCFunction)__pyx_pw_5numba_15extension_types_5compute_object_offsets, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_5numba_15extension_types_4compute_object_offsets)};
static PyObject *__pyx_pw_5numba_15extension_types_5compute_object_offsets(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ext_type = 0;
  PyObject *__pyx_v_struct_offset = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_object_offsets (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__ext_type,&__pyx_n_s__struct_offset,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__ext_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__struct_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_object_offsets", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 88; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_object_offsets") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 88; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_ext_type = values[0];
    __pyx_v_struct_offset = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_object_offsets", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 88; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("numba.extension_types.compute_object_offsets", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5numba_15extension_types_4compute_object_offsets(__pyx_self, __pyx_v_ext_type, __pyx_v_struct_offset);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numba/extension_types.pyx":88
 * # on the extension type object.
 * 
 * def compute_object_offsets(ext_type, struct_offset):             # <<<<<<<<<<<<<<
 *     """
 *     Given an typesystem.ExtensionType, compute the offsets of object
 */

static PyObject *__pyx_pf_5numba_15extension_types_4compute_object_offsets(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ext_type, PyObject *__pyx_v_struct_offset) {
  PyObject *__pyx_v_offsets = NULL;
  PyObject *__pyx_v_field_name = NULL;
  PyObject *__pyx_v_field_type = NULL;
  PyObject *__pyx_v_offset = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_object_offsets", 0);

  /* "numba/extension_types.pyx":93
 *     pointers, relative to the start of the attribute offsets.
 *     """
 *     offsets = []             # <<<<<<<<<<<<<<
 *     for field_name, field_type in ext_type.attribute_struct.fields:
 *         if field_type.is_object or field_type.is_array:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 93; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_offsets = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":94
 *     """
 *     offsets = []
 *     for field_name, field_type in ext_type.attribute_struct.fields:             # <<<<<<<<<<<<<<
 *         if field_type.is_object or field_type.is_array:
 *             offset = ext_type.attribute_struct.offsetof(field_name)
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_ext_type, __pyx_n_s__attribute_struct); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__fields); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyList_CheckExact(__pyx_t_2) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (!__pyx_t_4 && PyList_CheckExact(__pyx_t_1)) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_COMPILING_IN_CPYTHON
      __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      #else
      __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      #endif
    } else if (!__pyx_t_4 && PyTuple_CheckExact(__pyx_t_1)) {
      if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_COMPILING_IN_CPYTHON
      __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      #else
      __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      #endif
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        if (PyErr_Occurred()) {
          if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      #if CYTHON_COMPILING_IN_CPYTHON
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      #if CYTHON_COMPILING_IN_CPYTHON
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF(__pyx_v_field_name);
    __pyx_v_field_name = __pyx_t_5;
    __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_v_field_type);
    __pyx_v_field_type = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "numba/extension_types.pyx":95
 *     offsets = []
 *     for field_name, field_type in ext_type.attribute_struct.fields:
 *         if field_type.is_object or field_type.is_array:             # <<<<<<<<<<<<<<
 *             offset = ext_type.attribute_struct.offsetof(field_name)
 *             offsets.append(struct_offset + offset)
 */
    __pyx_t_2 = PyObject_GetAttr(__pyx_v_field_type, __pyx_n_s__is_object); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 95; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_9 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 95; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!__pyx_t_9) {
      __pyx_t_2 = PyObject_GetAttr(__pyx_v_field_type, __pyx_n_s__is_array); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 95; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_10 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 95; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = __pyx_t_10;
    } else {
      __pyx_t_11 = __pyx_t_9;
    }
    if (__pyx_t_11) {

      /* "numba/extension_types.pyx":96
 *     for field_name, field_type in ext_type.attribute_struct.fields:
 *         if field_type.is_object or field_type.is_array:
 *             offset = ext_type.attribute_struct.offsetof(field_name)             # <<<<<<<<<<<<<<
 *             offsets.append(struct_offset + offset)
 * 
 */
      __pyx_t_2 = PyObject_GetAttr(__pyx_v_ext_type, __pyx_n_s__attribute_struct); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 96; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__offsetof); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 96; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 96; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_field_name);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_field_name);
      __Pyx_GIVEREF(__pyx_v_field_name);
      __pyx_t_5 = PyObject_Call(__pyx_t_6, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 96; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_v_offset);
      __pyx_v_offset = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "numba/extension_types.pyx":97
 *         if field_type.is_object or field_type.is_array:
 *             offset = ext_type.attribute_struct.offsetof(field_name)
 *             offsets.append(struct_offset + offset)             # <<<<<<<<<<<<<<
 * 
 *     return offsets
 */
      __pyx_t_5 = PyNumber_Add(__pyx_v_struct_offset, __pyx_v_offset); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_offsets, __pyx_t_5); if (unlikely(__pyx_t_12 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L7;
    }
    __pyx_L7:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":99
 *             offsets.append(struct_offset + offset)
 * 
 *     return offsets             # <<<<<<<<<<<<<<
 * 
 * cdef inline list getoffsets(type_obj):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_offsets));
  __pyx_r = ((PyObject *)__pyx_v_offsets);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("numba.extension_types.compute_object_offsets", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_offsets);
  __Pyx_XDECREF(__pyx_v_field_name);
  __Pyx_XDECREF(__pyx_v_field_type);
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numba/extension_types.pyx":101
 *     return offsets
 * 
 * cdef inline list getoffsets(type_obj):             # <<<<<<<<<<<<<<
 *     cdef list offsets = type_obj.__numba_object_offset
 *     return offsets
 */

static CYTHON_INLINE PyObject *__pyx_f_5numba_15extension_types_getoffsets(PyObject *__pyx_v_type_obj) {
  PyObject *__pyx_v_offsets = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getoffsets", 0);

  /* "numba/extension_types.pyx":102
 * 
 * cdef inline list getoffsets(type_obj):
 *     cdef list offsets = type_obj.__numba_object_offset             # <<<<<<<<<<<<<<
 *     return offsets
 * 
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_type_obj, __pyx_n_s_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 102; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected list, got %.200s", Py_TYPE(__pyx_t_1)->tp_name), 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 102; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_offsets = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":103
 * cdef inline list getoffsets(type_obj):
 *     cdef list offsets = type_obj.__numba_object_offset
 *     return offsets             # <<<<<<<<<<<<<<
 * 
 * cdef inline void setnone(PyObject **slot):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_offsets));
  __pyx_r = __pyx_v_offsets;
  goto __pyx_L0;

  __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numba.extension_types.getoffsets", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_offsets);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numba/extension_types.pyx":105
 *     return offsets
 * 
 * cdef inline void setnone(PyObject **slot):             # <<<<<<<<<<<<<<
 *     cdef PyObject *tmp
 * 
 */

static CYTHON_INLINE void __pyx_f_5numba_15extension_types_setnone(PyObject **__pyx_v_slot) {
  PyObject *__pyx_v_tmp;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setnone", 0);

  /* "numba/extension_types.pyx":108
 *     cdef PyObject *tmp
 * 
 *     tmp = slot[0]             # <<<<<<<<<<<<<<
 *     Py_INCREF(<PyObject *> None)
 *     slot[0] = <PyObject *> None
 */
  __pyx_v_tmp = (__pyx_v_slot[0]);

  /* "numba/extension_types.pyx":109
 * 
 *     tmp = slot[0]
 *     Py_INCREF(<PyObject *> None)             # <<<<<<<<<<<<<<
 *     slot[0] = <PyObject *> None
 *     Py_XDECREF(tmp)
 */
  Py_INCREF(((PyObject *)Py_None));

  /* "numba/extension_types.pyx":110
 *     tmp = slot[0]
 *     Py_INCREF(<PyObject *> None)
 *     slot[0] = <PyObject *> None             # <<<<<<<<<<<<<<
 *     Py_XDECREF(tmp)
 * 
 */
  (__pyx_v_slot[0]) = ((PyObject *)Py_None);

  /* "numba/extension_types.pyx":111
 *     Py_INCREF(<PyObject *> None)
 *     slot[0] = <PyObject *> None
 *     Py_XDECREF(tmp)             # <<<<<<<<<<<<<<
 * 
 * cdef void tp_dealloc(PyObject *self):
 */
  Py_XDECREF(__pyx_v_tmp);

  __Pyx_RefNannyFinishContext();
}

/* "numba/extension_types.pyx":113
 *     Py_XDECREF(tmp)
 * 
 * cdef void tp_dealloc(PyObject *self):             # <<<<<<<<<<<<<<
 *     """
 *     Dealloc for numba extension types.
 */

static void __pyx_f_5numba_15extension_types_tp_dealloc(PyObject *__pyx_v_self) {
  PyObject *__pyx_v_type_obj = NULL;
  PyTypeObject *__pyx_v_ob_type;
  Py_ssize_t __pyx_v_offset;
  PyObject **__pyx_v_slot;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("tp_dealloc", 0);

  /* "numba/extension_types.pyx":121
 *     the deallocator of the supertype, which uses the offsets of the subtype.
 *     """
 *     type_obj = <object> self.ob_type             # <<<<<<<<<<<<<<
 *     cdef PyTypeObject *ob_type = <PyTypeObject *> type_obj
 *     cdef Py_ssize_t offset
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self->ob_type));
  __pyx_v_type_obj = ((PyObject *)__pyx_v_self->ob_type);

  /* "numba/extension_types.pyx":122
 *     """
 *     type_obj = <object> self.ob_type
 *     cdef PyTypeObject *ob_type = <PyTypeObject *> type_obj             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t offset
 *     cdef PyObject **slot
 */
  __pyx_v_ob_type = ((PyTypeObject *)__pyx_v_type_obj);

  /* "numba/extension_types.pyx":126
 *     cdef PyObject **slot
 * 
 *     for offset in getoffsets(type_obj):             # <<<<<<<<<<<<<<
 *         slot = <PyObject **> ((<char *> <PyObject *> self) + offset)
 *         # Py_CLEAR(slot[0])
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5numba_15extension_types_getoffsets(__pyx_v_type_obj)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 126; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 126; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 126; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 126; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #endif
    __pyx_t_4 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_4 == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 126; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_offset = __pyx_t_4;

    /* "numba/extension_types.pyx":127
 * 
 *     for offset in getoffsets(type_obj):
 *         slot = <PyObject **> ((<char *> <PyObject *> self) + offset)             # <<<<<<<<<<<<<<
 *         # Py_CLEAR(slot[0])
 *         setnone(slot) # Does ctypes collect and traverse for us?
 */
    __pyx_v_slot = ((PyObject **)(((char *)((PyObject *)__pyx_v_self)) + __pyx_v_offset));

    /* "numba/extension_types.pyx":129
 *         slot = <PyObject **> ((<char *> <PyObject *> self) + offset)
 *         # Py_CLEAR(slot[0])
 *         setnone(slot) # Does ctypes collect and traverse for us?             # <<<<<<<<<<<<<<
 * 
 *     # Call base class dealloc
 */
    __pyx_f_5numba_15extension_types_setnone(__pyx_v_slot);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "numba/extension_types.pyx":132
 * 
 *     # Call base class dealloc
 *     if ob_type.tp_base.tp_dealloc != NULL:             # <<<<<<<<<<<<<<
 *         ob_type.tp_base.tp_dealloc(<PyObject *> self)
 * 
 */
  __pyx_t_5 = (__pyx_v_ob_type->tp_base->tp_dealloc != NULL);
  if (__pyx_t_5) {

    /* "numba/extension_types.pyx":133
 *     # Call base class dealloc
 *     if ob_type.tp_base.tp_dealloc != NULL:
 *         ob_type.tp_base.tp_dealloc(<PyObject *> self)             # <<<<<<<<<<<<<<
 * 
 * cdef int tp_clear(PyObject *self):
 */
    __pyx_v_ob_type->tp_base->tp_dealloc(((PyObject *)__pyx_v_self));
    goto __pyx_L5;
  }
  __pyx_L5:;

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("numba.extension_types.tp_dealloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_type_obj);
  __Pyx_RefNannyFinishContext();
}

/* "numba/extension_types.pyx":135
 *         ob_type.tp_base.tp_dealloc(<PyObject *> self)
 * 
 * cdef int tp_clear(PyObject *self):             # <<<<<<<<<<<<<<
 *     """
 *     tp_clear for numba extension types.
 */

static int __pyx_f_5numba_15extension_types_tp_clear(PyObject *__pyx_v_self) {
  PyTypeObject *__pyx_v_ob_type;
  Py_ssize_t __pyx_v_offset;
  PyObject **__pyx_v_slot;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("tp_clear", 0);

  /* "numba/extension_types.pyx":139
 *     tp_clear for numba extension types.
 *     """
 *     cdef PyTypeObject *ob_type = <PyTypeObject *> self.ob_type             # <<<<<<<<<<<<<<
 * 
 *     # Let superclass clear its attributes
 */
  __pyx_v_ob_type = ((PyTypeObject *)__pyx_v_self->ob_type);

  /* "numba/extension_types.pyx":142
 * 
 *     # Let superclass clear its attributes
 *     if ob_type.tp_base.tp_clear != NULL:             # <<<<<<<<<<<<<<
 *         ob_type.tp_base.tp_clear(self)
 * 
 */
  __pyx_t_1 = (__pyx_v_ob_type->tp_base->tp_clear != NULL);
  if (__pyx_t_1) {

    /* "numba/extension_types.pyx":143
 *     # Let superclass clear its attributes
 *     if ob_type.tp_base.tp_clear != NULL:
 *         ob_type.tp_base.tp_clear(self)             # <<<<<<<<<<<<<<
 * 
 *     # Clear our own attributes
 */
    __pyx_v_ob_type->tp_base->tp_clear(__pyx_v_self);
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numba/extension_types.pyx":150
 * 
 * 
 *     for offset in getoffsets(<object> ob_type):             # <<<<<<<<<<<<<<
 *         slot = <PyObject **> ((<char *> <PyObject *> self) + offset)
 *         setnone(slot)
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_5numba_15extension_types_getoffsets(((PyObject *)__pyx_v_ob_type))); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 150; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 150; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 150; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #else
    __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 150; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #endif
    __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 150; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_offset = __pyx_t_5;

    /* "numba/extension_types.pyx":151
 * 
 *     for offset in getoffsets(<object> ob_type):
 *         slot = <PyObject **> ((<char *> <PyObject *> self) + offset)             # <<<<<<<<<<<<<<
 *         setnone(slot)
 * 
 */
    __pyx_v_slot = ((PyObject **)(((char *)((PyObject *)__pyx_v_self)) + __pyx_v_offset));

    /* "numba/extension_types.pyx":152
 *     for offset in getoffsets(<object> ob_type):
 *         slot = <PyObject **> ((<char *> <PyObject *> self) + offset)
 *         setnone(slot)             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    __pyx_f_5numba_15extension_types_setnone(__pyx_v_slot);
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "numba/extension_types.pyx":154
 *         setnone(slot)
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef int tp_traverse(PyObject *self, visitproc visit, void *arg):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("numba.extension_types.tp_clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numba/extension_types.pyx":156
 *     return 0
 * 
 * cdef int tp_traverse(PyObject *self, visitproc visit, void *arg):             # <<<<<<<<<<<<<<
 *     """
 *     tp_traverse for numba extension types.
 */

static int __pyx_f_5numba_15extension_types_tp_traverse(PyObject *__pyx_v_self, __pyx_t_5numba_15extension_types_visitproc __pyx_v_visit, void *__pyx_v_arg) {
  PyTypeObject *__pyx_v_ob_type;
  int __pyx_v_result;
  Py_ssize_t __pyx_v_offset;
  PyObject *__pyx_v_obj;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("tp_traverse", 0);

  /* "numba/extension_types.pyx":160
 *     tp_traverse for numba extension types.
 *     """
 *     cdef PyTypeObject *ob_type = <PyTypeObject *> self.ob_type             # <<<<<<<<<<<<<<
 *     cdef int result
 * 
 */
  __pyx_v_ob_type = ((PyTypeObject *)__pyx_v_self->ob_type);

  /* "numba/extension_types.pyx":163
 *     cdef int result
 * 
 *     if ob_type.tp_base.tp_traverse != NULL:             # <<<<<<<<<<<<<<
 *         result = ob_type.tp_base.tp_traverse(self, visit, arg)
 *         if result:
 */
  __pyx_t_1 = (__pyx_v_ob_type->tp_base->tp_traverse != NULL);
  if (__pyx_t_1) {

    /* "numba/extension_types.pyx":164
 * 
 *     if ob_type.tp_base.tp_traverse != NULL:
 *         result = ob_type.tp_base.tp_traverse(self, visit, arg)             # <<<<<<<<<<<<<<
 *         if result:
 *             return result
 */
    __pyx_v_result = __pyx_v_ob_type->tp_base->tp_traverse(__pyx_v_self, __pyx_v_visit, __pyx_v_arg);

    /* "numba/extension_types.pyx":165
 *     if ob_type.tp_base.tp_traverse != NULL:
 *         result = ob_type.tp_base.tp_traverse(self, visit, arg)
 *         if result:             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
    if (__pyx_v_result) {

      /* "numba/extension_types.pyx":166
 *         result = ob_type.tp_base.tp_traverse(self, visit, arg)
 *         if result:
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t offset
 */
      __pyx_r = __pyx_v_result;
      goto __pyx_L0;
      goto __pyx_L4;
    }
    __pyx_L4:;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numba/extension_types.pyx":171
 *     cdef PyObject *obj
 * 
 *     for offset in getoffsets(<object> ob_type):             # <<<<<<<<<<<<<<
 *         obj = (<PyObject **> ((<char *> <PyObject *> self) + offset))[0]
 *         if obj != NULL:
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_5numba_15extension_types_getoffsets(((PyObject *)__pyx_v_ob_type))); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 171; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 171; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 171; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #else
    __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 171; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #endif
    __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 171; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_offset = __pyx_t_5;

    /* "numba/extension_types.pyx":172
 * 
 *     for offset in getoffsets(<object> ob_type):
 *         obj = (<PyObject **> ((<char *> <PyObject *> self) + offset))[0]             # <<<<<<<<<<<<<<
 *         if obj != NULL:
 *             result = visit(obj, arg)
 */
    __pyx_v_obj = (((PyObject **)(((char *)((PyObject *)__pyx_v_self)) + __pyx_v_offset))[0]);

    /* "numba/extension_types.pyx":173
 *     for offset in getoffsets(<object> ob_type):
 *         obj = (<PyObject **> ((<char *> <PyObject *> self) + offset))[0]
 *         if obj != NULL:             # <<<<<<<<<<<<<<
 *             result = visit(obj, arg)
 *             if result:
 */
    __pyx_t_1 = (__pyx_v_obj != NULL);
    if (__pyx_t_1) {

      /* "numba/extension_types.pyx":174
 *         obj = (<PyObject **> ((<char *> <PyObject *> self) + offset))[0]
 *         if obj != NULL:
 *             result = visit(obj, arg)             # <<<<<<<<<<<<<<
 *             if result:
 *                 return result
 */
      __pyx_v_result = __pyx_v_visit(__pyx_v_obj, __pyx_v_arg);

      /* "numba/extension_types.pyx":175
 *         if obj != NULL:
 *             result = visit(obj, arg)
 *             if result:             # <<<<<<<<<<<<<<
 *                 return result
 * 
 */
      if (__pyx_v_result) {

        /* "numba/extension_types.pyx":176
 *             result = visit(obj, arg)
 *             if result:
 *                 return result             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
        __pyx_r = __pyx_v_result;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L0;
        goto __pyx_L8;
      }
      __pyx_L8:;
      goto __pyx_L7;
    }
    __pyx_L7:;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "numba/extension_types.pyx":178
 *                 return result
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * #------------------------------------------------------------------------
 */
  __pyx_r = 0;
  goto __pyx_L0;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("numba.extension_types.tp_traverse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5numba_15extension_types_7create_new_extension_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5numba_15extension_types_6create_new_extension_type[] = "\n    Create an extension type from the given name, bases and dict. Also\n    takes a vtab struct minitype, and a struct_type describing the\n    object attributes.\n    ";
static PyMethodDef __pyx_mdef_5numba_15extension_types_7create_new_extension_type = {__Pyx_NAMESTR("create_new_extension_type"), (PyCFunction)__pyx_pw_5numba_15extension_types_7create_new_extension_type, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_5numba_15extension_types_6create_new_extension_type)};
static PyObject *__pyx_pw_5numba_15extension_types_7create_new_extension_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_bases = 0;
  PyObject *__pyx_v_dict = 0;
  PyObject *__pyx_v_ext_numba_type = 0;
  PyObject *__pyx_v_vtab = 0;
  PyObject *__pyx_v_vtab_type = 0;
  PyObject *__pyx_v_llvm_methods = 0;
  PyObject *__pyx_v_method_pointers = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_new_extension_type (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__name,&__pyx_n_s__bases,&__pyx_n_s__dict,&__pyx_n_s__ext_numba_type,&__pyx_n_s__vtab,&__pyx_n_s__vtab_type,&__pyx_n_s__llvm_methods,&__pyx_n_s__method_pointers,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__bases)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_new_extension_type", 1, 8, 8, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_new_extension_type", 1, 8, 8, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__ext_numba_type)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_new_extension_type", 1, 8, 8, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__vtab)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_new_extension_type", 1, 8, 8, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__vtab_type)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_new_extension_type", 1, 8, 8, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__llvm_methods)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_new_extension_type", 1, 8, 8, 6); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__method_pointers)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_new_extension_type", 1, 8, 8, 7); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_new_extension_type") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_name = values[0];
    __pyx_v_bases = values[1];
    __pyx_v_dict = values[2];
    __pyx_v_ext_numba_type = values[3];
    __pyx_v_vtab = values[4];
    __pyx_v_vtab_type = values[5];
    __pyx_v_llvm_methods = values[6];
    __pyx_v_method_pointers = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_new_extension_type", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("numba.extension_types.create_new_extension_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5numba_15extension_types_6create_new_extension_type(__pyx_self, __pyx_v_name, __pyx_v_bases, __pyx_v_dict, __pyx_v_ext_numba_type, __pyx_v_vtab, __pyx_v_vtab_type, __pyx_v_llvm_methods, __pyx_v_method_pointers);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5numba_15extension_types_25create_new_extension_type_1new(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5numba_15extension_types_25create_new_extension_type_new[] = "Create a new object and patch it with a vtab";
static PyMethodDef __pyx_mdef_5numba_15extension_types_25create_new_extension_type_1new = {__Pyx_NAMESTR("new"), (PyCFunction)__pyx_pw_5numba_15extension_types_25create_new_extension_type_1new, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_5numba_15extension_types_25create_new_extension_type_new)};
static PyObject *__pyx_pw_5numba_15extension_types_25create_new_extension_type_1new(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwds = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("new (wrapper)", 0);
  __pyx_v_kwds = PyDict_New(); if (unlikely(!__pyx_v_kwds)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwds);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwds); __pyx_v_kwds = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__cls,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwds, values, used_pos_args, "new") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_cls = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("new", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwds); __pyx_v_kwds = 0;
  __Pyx_AddTraceback("numba.extension_types.create_new_extension_type.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5numba_15extension_types_25create_new_extension_type_new(__pyx_self, __pyx_v_cls, __pyx_v_args, __pyx_v_kwds);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwds);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numba/extension_types.pyx":195
 * 
 *     orig_new = dict.get('__new__', None)
 *     def new(cls, *args, **kwds):             # <<<<<<<<<<<<<<
 *         "Create a new object and patch it with a vtab"
 *         cdef PyObject *obj_p
 */

static PyObject *__pyx_pf_5numba_15extension_types_25create_new_extension_type_new(PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_args, PyObject *__pyx_v_kwds) {
  struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *__pyx_cur_scope;
  struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *__pyx_outer_scope;
  PyObject *__pyx_v_obj_p;
  void **__pyx_v_vtab_location;
  PyObject *__pyx_v_new_func = NULL;
  PyObject *__pyx_v_obj = NULL;
  Py_ssize_t __pyx_v_attrs_pointer;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_uintptr_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 0);
  __pyx_outer_scope = (struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "numba/extension_types.pyx":200
 *         cdef void **vtab_location
 * 
 *         if orig_new is not None:             # <<<<<<<<<<<<<<
 *             new_func = orig_new
 *         else:
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_orig_new)) { __Pyx_RaiseClosureNameError("orig_new"); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 200; __pyx_clineno = __LINE__; goto __pyx_L1_error;} }
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_orig_new != Py_None);
  if (__pyx_t_1) {

    /* "numba/extension_types.pyx":201
 * 
 *         if orig_new is not None:
 *             new_func = orig_new             # <<<<<<<<<<<<<<
 *         else:
 *             assert issubclass(cls, ext_type), (cls, ext_type)
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_orig_new);
    __pyx_v_new_func = __pyx_cur_scope->__pyx_v_orig_new;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "numba/extension_types.pyx":203
 *             new_func = orig_new
 *         else:
 *             assert issubclass(cls, ext_type), (cls, ext_type)             # <<<<<<<<<<<<<<
 *             new_func = super(ext_type, cls).__new__
 * 
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!__pyx_cur_scope->__pyx_v_ext_type)) { __Pyx_RaiseClosureNameError("ext_type"); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 203; __pyx_clineno = __LINE__; goto __pyx_L1_error;} }
    __pyx_t_2 = ((PyObject *)__pyx_cur_scope->__pyx_v_ext_type);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = PyObject_IsSubclass(__pyx_v_cls, __pyx_t_2); if (unlikely(__pyx_t_1 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 203; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) {
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 203; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_cls);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_cls);
      __Pyx_GIVEREF(__pyx_v_cls);
      __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
      PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
      __Pyx_GIVEREF(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
      PyErr_SetObject(PyExc_AssertionError, ((PyObject *)__pyx_t_2));
      __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 203; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    #endif

    /* "numba/extension_types.pyx":204
 *         else:
 *             assert issubclass(cls, ext_type), (cls, ext_type)
 *             new_func = super(ext_type, cls).__new__             # <<<<<<<<<<<<<<
 * 
 *         if base_is_object:
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 204; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
    __Pyx_GIVEREF(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
    __Pyx_INCREF(__pyx_v_cls);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_cls);
    __Pyx_GIVEREF(__pyx_v_cls);
    __pyx_t_3 = PyObject_Call(__pyx_builtin_super, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 204; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s____new__); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 204; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_new_func = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "numba/extension_types.pyx":206
 *             new_func = super(ext_type, cls).__new__
 * 
 *         if base_is_object:             # <<<<<<<<<<<<<<
 *             # Avoid warnings in py2.6:
 *             #     DeprecationWarning: object.__new__() takes no parameters
 */
  if (__pyx_cur_scope->__pyx_v_base_is_object) {

    /* "numba/extension_types.pyx":209
 *             # Avoid warnings in py2.6:
 *             #     DeprecationWarning: object.__new__() takes no parameters
 *             obj = new_func(cls)             # <<<<<<<<<<<<<<
 *         else:
 *             obj = new_func(cls, *args, **kwds)
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 209; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_cls);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_cls);
    __Pyx_GIVEREF(__pyx_v_cls);
    __pyx_t_3 = PyObject_Call(__pyx_v_new_func, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 209; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
    __pyx_v_obj = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L4;
  }
  /*else*/ {

    /* "numba/extension_types.pyx":211
 *             obj = new_func(cls)
 *         else:
 *             obj = new_func(cls, *args, **kwds)             # <<<<<<<<<<<<<<
 * 
 *         if (cls.__numba_vtab is not ext_type.__numba_vtab or
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 211; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_cls);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_cls);
    __Pyx_GIVEREF(__pyx_v_cls);
    __pyx_t_2 = PySequence_Tuple(((PyObject *)__pyx_v_args)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 211; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_2));
    __pyx_t_4 = PyNumber_Add(((PyObject *)__pyx_t_3), ((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 211; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_4));
    __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
    __pyx_t_2 = ((PyObject *)__pyx_v_kwds);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = PyObject_Call(__pyx_v_new_func, ((PyObject *)__pyx_t_4), ((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 211; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
    __pyx_v_obj = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L4:;

  /* "numba/extension_types.pyx":213
 *             obj = new_func(cls, *args, **kwds)
 * 
 *         if (cls.__numba_vtab is not ext_type.__numba_vtab or             # <<<<<<<<<<<<<<
 *                 not isinstance(obj, cls)):
 *             # Subclass will set the vtab and attributes
 */
  __pyx_t_3 = PyObject_GetAttr(__pyx_v_cls, __pyx_n_s____numba_vtab); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 213; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_cur_scope->__pyx_v_ext_type)) { __Pyx_RaiseClosureNameError("ext_type"); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 213; __pyx_clineno = __LINE__; goto __pyx_L1_error;} }
  __pyx_t_2 = PyObject_GetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s____numba_vtab); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 213; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = (__pyx_t_3 != __pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_1) {

    /* "numba/extension_types.pyx":214
 * 
 *         if (cls.__numba_vtab is not ext_type.__numba_vtab or
 *                 not isinstance(obj, cls)):             # <<<<<<<<<<<<<<
 *             # Subclass will set the vtab and attributes
 *             return obj
 */
    __pyx_t_5 = PyObject_IsInstance(__pyx_v_obj, __pyx_v_cls); if (unlikely(__pyx_t_5 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 214; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_6 = (!__pyx_t_5);
    __pyx_t_5 = __pyx_t_6;
  } else {
    __pyx_t_5 = __pyx_t_1;
  }
  if (__pyx_t_5) {

    /* "numba/extension_types.pyx":216
 *                 not isinstance(obj, cls)):
 *             # Subclass will set the vtab and attributes
 *             return obj             # <<<<<<<<<<<<<<
 * 
 *         # It is our responsibility to set the vtab and the ctypes attributes
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_obj);
    __pyx_r = __pyx_v_obj;
    goto __pyx_L0;
    goto __pyx_L5;
  }
  __pyx_L5:;

  /* "numba/extension_types.pyx":220
 *         # It is our responsibility to set the vtab and the ctypes attributes
 *         # Other fields are 0/NULL
 *         obj_p = <PyObject *> obj             # <<<<<<<<<<<<<<
 * 
 *         vtab_location = <void **> ((<char *> obj_p) + vtab_offset)
 */
  __pyx_v_obj_p = ((PyObject *)__pyx_v_obj);

  /* "numba/extension_types.pyx":222
 *         obj_p = <PyObject *> obj
 * 
 *         vtab_location = <void **> ((<char *> obj_p) + vtab_offset)             # <<<<<<<<<<<<<<
 *         if vtab:
 *             vtab_location[0] = <void *> <Py_uintptr_t> cls.__numba_vtab_p
 */
  __pyx_v_vtab_location = ((void **)(((char *)__pyx_v_obj_p) + __pyx_cur_scope->__pyx_v_vtab_offset));

  /* "numba/extension_types.pyx":223
 * 
 *         vtab_location = <void **> ((<char *> obj_p) + vtab_offset)
 *         if vtab:             # <<<<<<<<<<<<<<
 *             vtab_location[0] = <void *> <Py_uintptr_t> cls.__numba_vtab_p
 *         else:
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_vtab)) { __Pyx_RaiseClosureNameError("vtab"); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 223; __pyx_clineno = __LINE__; goto __pyx_L1_error;} }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_vtab); if (unlikely(__pyx_t_5 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 223; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_5) {

    /* "numba/extension_types.pyx":224
 *         vtab_location = <void **> ((<char *> obj_p) + vtab_offset)
 *         if vtab:
 *             vtab_location[0] = <void *> <Py_uintptr_t> cls.__numba_vtab_p             # <<<<<<<<<<<<<<
 *         else:
 *             vtab_location[0] = NULL
 */
    __pyx_t_2 = PyObject_GetAttr(__pyx_v_cls, __pyx_n_s____numba_vtab_p); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 224; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyInt_from_py_Py_uintptr_t(__pyx_t_2); if (unlikely((__pyx_t_7 == (Py_uintptr_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 224; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_vtab_location[0]) = ((void *)((Py_uintptr_t)__pyx_t_7));
    goto __pyx_L6;
  }
  /*else*/ {

    /* "numba/extension_types.pyx":226
 *             vtab_location[0] = <void *> <Py_uintptr_t> cls.__numba_vtab_p
 *         else:
 *             vtab_location[0] = NULL             # <<<<<<<<<<<<<<
 * 
 *         attrs_pointer = (<Py_uintptr_t> obj_p) + attrs_offset
 */
    (__pyx_v_vtab_location[0]) = NULL;
  }
  __pyx_L6:;

  /* "numba/extension_types.pyx":228
 *             vtab_location[0] = NULL
 * 
 *         attrs_pointer = (<Py_uintptr_t> obj_p) + attrs_offset             # <<<<<<<<<<<<<<
 *         obj._numba_attrs = ctypes.cast(attrs_pointer,
 *                                        cls.__numba_struct_ctype_p)[0]
 */
  __pyx_v_attrs_pointer = (((Py_uintptr_t)__pyx_v_obj_p) + __pyx_cur_scope->__pyx_v_attrs_offset);

  /* "numba/extension_types.pyx":229
 * 
 *         attrs_pointer = (<Py_uintptr_t> obj_p) + attrs_offset
 *         obj._numba_attrs = ctypes.cast(attrs_pointer,             # <<<<<<<<<<<<<<
 *                                        cls.__numba_struct_ctype_p)[0]
 * 
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__ctypes); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__cast); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_attrs_pointer); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);

  /* "numba/extension_types.pyx":230
 *         attrs_pointer = (<Py_uintptr_t> obj_p) + attrs_offset
 *         obj._numba_attrs = ctypes.cast(attrs_pointer,
 *                                        cls.__numba_struct_ctype_p)[0]             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
  __pyx_t_4 = PyObject_GetAttr(__pyx_v_cls, __pyx_n_s_2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 230; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_8), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_8)); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_4, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_8) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 230; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "numba/extension_types.pyx":229
 * 
 *         attrs_pointer = (<Py_uintptr_t> obj_p) + attrs_offset
 *         obj._numba_attrs = ctypes.cast(attrs_pointer,             # <<<<<<<<<<<<<<
 *                                        cls.__numba_struct_ctype_p)[0]
 * 
 */
  if (PyObject_SetAttr(__pyx_v_obj, __pyx_n_s___numba_attrs, __pyx_t_8) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "numba/extension_types.pyx":232
 *                                        cls.__numba_struct_ctype_p)[0]
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     dict['__new__'] = staticmethod(new)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numba.extension_types.create_new_extension_type.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_new_func);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numba/extension_types.pyx":184
 * #------------------------------------------------------------------------
 * 
 * def create_new_extension_type(name, bases, dict, ext_numba_type,             # <<<<<<<<<<<<<<
 *                               vtab, vtab_type, llvm_methods, method_pointers):
 *     """
 */

static PyObject *__pyx_pf_5numba_15extension_types_6create_new_extension_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_bases, PyObject *__pyx_v_dict, PyObject *__pyx_v_ext_numba_type, PyObject *__pyx_v_vtab, PyObject *__pyx_v_vtab_type, PyObject *__pyx_v_llvm_methods, PyObject *__pyx_v_method_pointers) {
  struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *__pyx_cur_scope;
  PyTypeObject *__pyx_v_ext_type_p;
  PyObject *__pyx_v_new = 0;
  PyObject *__pyx_v_struct_type = NULL;
  PyObject *__pyx_v_struct_ctype = NULL;
  PyObject *__pyx_v_upper = NULL;
  PyObject *__pyx_v_vtab_p = NULL;
  PyObject *__pyx_v_offsets = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_uintptr_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_new_extension_type", 0);
  __pyx_cur_scope = (struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *)__pyx_ptype_5numba_15extension_types___pyx_scope_struct__create_new_extension_type->tp_new(__pyx_ptype_5numba_15extension_types___pyx_scope_struct__create_new_extension_type, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_cur_scope);
  __pyx_cur_scope->__pyx_v_vtab = __pyx_v_vtab;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_vtab);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_vtab);

  /* "numba/extension_types.pyx":194
 *     cdef Py_ssize_t vtab_offset, attrs_offset
 * 
 *     orig_new = dict.get('__new__', None)             # <<<<<<<<<<<<<<
 *     def new(cls, *args, **kwds):
 *         "Create a new object and patch it with a vtab"
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_dict, __pyx_n_s__get); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_k_tuple_3), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_orig_new = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "numba/extension_types.pyx":195
 * 
 *     orig_new = dict.get('__new__', None)
 *     def new(cls, *args, **kwds):             # <<<<<<<<<<<<<<
 *         "Create a new object and patch it with a vtab"
 *         cdef PyObject *obj_p
 */
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_5numba_15extension_types_25create_new_extension_type_1new, 0, __pyx_n_s_7, ((PyObject*)__pyx_cur_scope), __pyx_n_s_8, ((PyObject *)__pyx_k_codeobj_5)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_new = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "numba/extension_types.pyx":234
 *         return obj
 * 
 *     dict['__new__'] = staticmethod(new)             # <<<<<<<<<<<<<<
 *     ext_type = type(name, bases, dict)
 *     assert isinstance(ext_type, type)
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 234; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_new);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_new);
  __Pyx_GIVEREF(__pyx_v_new);
  __pyx_t_1 = PyObject_Call(__pyx_builtin_staticmethod, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 234; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  if (PyObject_SetItem(__pyx_v_dict, ((PyObject *)__pyx_n_s____new__), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 234; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":235
 * 
 *     dict['__new__'] = staticmethod(new)
 *     ext_type = type(name, bases, dict)             # <<<<<<<<<<<<<<
 *     assert isinstance(ext_type, type)
 * 
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 235; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_INCREF(__pyx_v_bases);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_bases);
  __Pyx_GIVEREF(__pyx_v_bases);
  __Pyx_INCREF(__pyx_v_dict);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_dict);
  __Pyx_GIVEREF(__pyx_v_dict);
  __pyx_t_2 = PyObject_Call(((PyObject *)((PyObject*)(&PyType_Type))), ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 235; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_ext_type = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "numba/extension_types.pyx":236
 *     dict['__new__'] = staticmethod(new)
 *     ext_type = type(name, bases, dict)
 *     assert isinstance(ext_type, type)             # <<<<<<<<<<<<<<
 * 
 *     cdef bint base_is_object = bases == (object,)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  __pyx_t_2 = ((PyObject *)__pyx_cur_scope->__pyx_v_ext_type);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_t_2, ((PyObject*)(&PyType_Type))); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) {
    PyErr_SetNone(PyExc_AssertionError);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 236; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  #endif

  /* "numba/extension_types.pyx":238
 *     assert isinstance(ext_type, type)
 * 
 *     cdef bint base_is_object = bases == (object,)             # <<<<<<<<<<<<<<
 * 
 *     ext_type_p = <PyTypeObject *> ext_type
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_builtin_object);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_builtin_object);
  __Pyx_GIVEREF(__pyx_builtin_object);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_bases, ((PyObject *)__pyx_t_2), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_cur_scope->__pyx_v_base_is_object = __pyx_t_3;

  /* "numba/extension_types.pyx":240
 *     cdef bint base_is_object = bases == (object,)
 * 
 *     ext_type_p = <PyTypeObject *> ext_type             # <<<<<<<<<<<<<<
 * 
 *     # Object offset for vtab is lower
 */
  __pyx_v_ext_type_p = ((PyTypeObject *)__pyx_cur_scope->__pyx_v_ext_type);

  /* "numba/extension_types.pyx":245
 *     # Object attributes are located at lower + sizeof(void *), and end at
 *     # upper
 *     struct_type = ext_numba_type.attribute_struct             # <<<<<<<<<<<<<<
 *     struct_ctype = struct_type.to_ctypes()
 *     vtab_offset = compute_vtab_offset(ext_type)
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_ext_numba_type, __pyx_n_s__attribute_struct); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct_type = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":246
 *     # upper
 *     struct_type = ext_numba_type.attribute_struct
 *     struct_ctype = struct_type.to_ctypes()             # <<<<<<<<<<<<<<
 *     vtab_offset = compute_vtab_offset(ext_type)
 *     attrs_offset = compute_attrs_offset(ext_type)
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_struct_type, __pyx_n_s__to_ctypes); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 246; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 246; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_struct_ctype = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "numba/extension_types.pyx":247
 *     struct_type = ext_numba_type.attribute_struct
 *     struct_ctype = struct_type.to_ctypes()
 *     vtab_offset = compute_vtab_offset(ext_type)             # <<<<<<<<<<<<<<
 *     attrs_offset = compute_attrs_offset(ext_type)
 *     upper = attrs_offset + ctypes.sizeof(struct_ctype)
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__compute_vtab_offset); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
  __Pyx_GIVEREF(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
  __pyx_t_4 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_cur_scope->__pyx_v_vtab_offset = __pyx_t_5;

  /* "numba/extension_types.pyx":248
 *     struct_ctype = struct_type.to_ctypes()
 *     vtab_offset = compute_vtab_offset(ext_type)
 *     attrs_offset = compute_attrs_offset(ext_type)             # <<<<<<<<<<<<<<
 *     upper = attrs_offset + ctypes.sizeof(struct_ctype)
 *     upper = align(upper, 8)
 */
  __pyx_t_4 = __Pyx_GetName(__pyx_m, __pyx_n_s_9); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
  __Pyx_GIVEREF(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
  __pyx_t_2 = PyObject_Call(__pyx_t_4, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_cur_scope->__pyx_v_attrs_offset = __pyx_t_5;

  /* "numba/extension_types.pyx":249
 *     vtab_offset = compute_vtab_offset(ext_type)
 *     attrs_offset = compute_attrs_offset(ext_type)
 *     upper = attrs_offset + ctypes.sizeof(struct_ctype)             # <<<<<<<<<<<<<<
 *     upper = align(upper, 8)
 * 
 */
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_cur_scope->__pyx_v_attrs_offset); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__ctypes); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__sizeof); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_struct_ctype);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_struct_ctype);
  __Pyx_GIVEREF(__pyx_v_struct_ctype);
  __pyx_t_6 = PyObject_Call(__pyx_t_4, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_upper = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":250
 *     attrs_offset = compute_attrs_offset(ext_type)
 *     upper = attrs_offset + ctypes.sizeof(struct_ctype)
 *     upper = align(upper, 8)             # <<<<<<<<<<<<<<
 * 
 *     # print 'basicsize/vtab_offset/upper', ext_type_p.tp_basicsize, lower, upper
 */
  __pyx_t_7 = __Pyx_PyInt_from_py_Py_uintptr_t(__pyx_v_upper); if (unlikely((__pyx_t_7 == (Py_uintptr_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_uintptr_t(__pyx_f_5numba_15extension_types_align(__pyx_t_7, 8)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_v_upper);
  __pyx_v_upper = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":253
 * 
 *     # print 'basicsize/vtab_offset/upper', ext_type_p.tp_basicsize, lower, upper
 *     ext_type_p.tp_basicsize = upper             # <<<<<<<<<<<<<<
 *     if ext_type_p.tp_itemsize:
 *         raise NotImplementedError("Subclassing variable sized objects")
 */
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_v_upper); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 253; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_ext_type_p->tp_basicsize = __pyx_t_5;

  /* "numba/extension_types.pyx":254
 *     # print 'basicsize/vtab_offset/upper', ext_type_p.tp_basicsize, lower, upper
 *     ext_type_p.tp_basicsize = upper
 *     if ext_type_p.tp_itemsize:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("Subclassing variable sized objects")
 * 
 */
  if (__pyx_v_ext_type_p->tp_itemsize) {

    /* "numba/extension_types.pyx":255
 *     ext_type_p.tp_basicsize = upper
 *     if ext_type_p.tp_itemsize:
 *         raise NotImplementedError("Subclassing variable sized objects")             # <<<<<<<<<<<<<<
 * 
 *     # TODO: Figure out ctypes policy on tp_dealloc/tp_traverse/tp_clear
 */
    __pyx_t_1 = PyObject_Call(__pyx_builtin_NotImplementedError, ((PyObject *)__pyx_k_tuple_11), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 255; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 255; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numba/extension_types.pyx":264
 *     #    ext_type_p.tp_traverse = tp_traverse
 * 
 *     ext_type.__numba_vtab_offset = vtab_offset             # <<<<<<<<<<<<<<
 *     ext_type.__numba_obj_end = upper
 *     ext_type.__numba_attr_offset = attrs_offset
 */
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_cur_scope->__pyx_v_vtab_offset); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 264; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s____numba_vtab_offset, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 264; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":265
 * 
 *     ext_type.__numba_vtab_offset = vtab_offset
 *     ext_type.__numba_obj_end = upper             # <<<<<<<<<<<<<<
 *     ext_type.__numba_attr_offset = attrs_offset
 * 
 */
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s____numba_obj_end, __pyx_v_upper) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 265; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":266
 *     ext_type.__numba_vtab_offset = vtab_offset
 *     ext_type.__numba_obj_end = upper
 *     ext_type.__numba_attr_offset = attrs_offset             # <<<<<<<<<<<<<<
 * 
 *     ext_type.__numba_vtab = vtab
 */
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_cur_scope->__pyx_v_attrs_offset); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 266; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s____numba_attr_offset, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 266; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":268
 *     ext_type.__numba_attr_offset = attrs_offset
 * 
 *     ext_type.__numba_vtab = vtab             # <<<<<<<<<<<<<<
 *     ext_type.__numba_vtab_type = vtab_type
 * 
 */
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s____numba_vtab, __pyx_cur_scope->__pyx_v_vtab) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 268; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":269
 * 
 *     ext_type.__numba_vtab = vtab
 *     ext_type.__numba_vtab_type = vtab_type             # <<<<<<<<<<<<<<
 * 
 *     if vtab:
 */
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s____numba_vtab_type, __pyx_v_vtab_type) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 269; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":271
 *     ext_type.__numba_vtab_type = vtab_type
 * 
 *     if vtab:             # <<<<<<<<<<<<<<
 *         vtab_p = ctypes.byref(vtab)
 *         ext_type.__numba_vtab_p = ctypes.cast(vtab_p, ctypes.c_void_p).value
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_vtab); if (unlikely(__pyx_t_3 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 271; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_3) {

    /* "numba/extension_types.pyx":272
 * 
 *     if vtab:
 *         vtab_p = ctypes.byref(vtab)             # <<<<<<<<<<<<<<
 *         ext_type.__numba_vtab_p = ctypes.cast(vtab_p, ctypes.c_void_p).value
 *     else:
 */
    __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__ctypes); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__byref); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_vtab);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_cur_scope->__pyx_v_vtab);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_vtab);
    __pyx_t_2 = PyObject_Call(__pyx_t_6, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
    __pyx_v_vtab_p = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "numba/extension_types.pyx":273
 *     if vtab:
 *         vtab_p = ctypes.byref(vtab)
 *         ext_type.__numba_vtab_p = ctypes.cast(vtab_p, ctypes.c_void_p).value             # <<<<<<<<<<<<<<
 *     else:
 *         ext_type.__numba_vtab_p = None
 */
    __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__ctypes); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__cast); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__ctypes); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__c_void_p); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_vtab_p);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_vtab_p);
    __Pyx_GIVEREF(__pyx_v_vtab_p);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_GetAttr(__pyx_t_6, __pyx_n_s__value); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s____numba_vtab_p, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L4;
  }
  /*else*/ {

    /* "numba/extension_types.pyx":275
 *         ext_type.__numba_vtab_p = ctypes.cast(vtab_p, ctypes.c_void_p).value
 *     else:
 *         ext_type.__numba_vtab_p = None             # <<<<<<<<<<<<<<
 * 
 *     ext_type.__numba_orig_tp_new = <Py_uintptr_t> ext_type_p.tp_new
 */
    if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s____numba_vtab_p, Py_None) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 275; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_L4:;

  /* "numba/extension_types.pyx":277
 *         ext_type.__numba_vtab_p = None
 * 
 *     ext_type.__numba_orig_tp_new = <Py_uintptr_t> ext_type_p.tp_new             # <<<<<<<<<<<<<<
 *     ext_type.__numba_struct_type = struct_type
 *     ext_type.__numba_struct_ctype_p = struct_type.pointer().to_ctypes()
 */
  __pyx_t_2 = __Pyx_PyInt_to_py_Py_uintptr_t(((Py_uintptr_t)__pyx_v_ext_type_p->tp_new)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s____numba_orig_tp_new, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "numba/extension_types.pyx":278
 * 
 *     ext_type.__numba_orig_tp_new = <Py_uintptr_t> ext_type_p.tp_new
 *     ext_type.__numba_struct_type = struct_type             # <<<<<<<<<<<<<<
 *     ext_type.__numba_struct_ctype_p = struct_type.pointer().to_ctypes()
 *     ext_type.__numba_lfuncs = llvm_methods
 */
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s____numba_struct_type, __pyx_v_struct_type) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 278; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":279
 *     ext_type.__numba_orig_tp_new = <Py_uintptr_t> ext_type_p.tp_new
 *     ext_type.__numba_struct_type = struct_type
 *     ext_type.__numba_struct_ctype_p = struct_type.pointer().to_ctypes()             # <<<<<<<<<<<<<<
 *     ext_type.__numba_lfuncs = llvm_methods
 *     ext_type.__numba_method_pointers = method_pointers
 */
  __pyx_t_2 = PyObject_GetAttr(__pyx_v_struct_type, __pyx_n_s__pointer); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_6, __pyx_n_s__to_ctypes); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s_2, __pyx_t_6) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "numba/extension_types.pyx":280
 *     ext_type.__numba_struct_type = struct_type
 *     ext_type.__numba_struct_ctype_p = struct_type.pointer().to_ctypes()
 *     ext_type.__numba_lfuncs = llvm_methods             # <<<<<<<<<<<<<<
 *     ext_type.__numba_method_pointers = method_pointers
 *     ext_type.__numba_ext_type = ext_numba_type
 */
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s____numba_lfuncs, __pyx_v_llvm_methods) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 280; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":281
 *     ext_type.__numba_struct_ctype_p = struct_type.pointer().to_ctypes()
 *     ext_type.__numba_lfuncs = llvm_methods
 *     ext_type.__numba_method_pointers = method_pointers             # <<<<<<<<<<<<<<
 *     ext_type.__numba_ext_type = ext_numba_type
 *     ext_type.exttype = ext_numba_type
 */
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s_12, __pyx_v_method_pointers) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 281; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":282
 *     ext_type.__numba_lfuncs = llvm_methods
 *     ext_type.__numba_method_pointers = method_pointers
 *     ext_type.__numba_ext_type = ext_numba_type             # <<<<<<<<<<<<<<
 *     ext_type.exttype = ext_numba_type
 * 
 */
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s____numba_ext_type, __pyx_v_ext_numba_type) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 282; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":283
 *     ext_type.__numba_method_pointers = method_pointers
 *     ext_type.__numba_ext_type = ext_numba_type
 *     ext_type.exttype = ext_numba_type             # <<<<<<<<<<<<<<
 * 
 *     offsets = getattr(ext_type, '__numba_object_offset', [])
 */
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s__exttype, __pyx_v_ext_numba_type) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 283; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":285
 *     ext_type.exttype = ext_numba_type
 * 
 *     offsets = getattr(ext_type, '__numba_object_offset', [])             # <<<<<<<<<<<<<<
 *     offsets = offsets + compute_object_offsets(ext_numba_type, attrs_offset)
 *     ext_type.__numba_object_offset = offsets
 */
  __pyx_t_6 = ((PyObject *)__pyx_cur_scope->__pyx_v_ext_type);
  __Pyx_INCREF(__pyx_t_6);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 285; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_t_6, ((PyObject *)__pyx_n_s_1), ((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 285; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_v_offsets = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":286
 * 
 *     offsets = getattr(ext_type, '__numba_object_offset', [])
 *     offsets = offsets + compute_object_offsets(ext_numba_type, attrs_offset)             # <<<<<<<<<<<<<<
 *     ext_type.__numba_object_offset = offsets
 * 
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s_13); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 286; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_cur_scope->__pyx_v_attrs_offset); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 286; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 286; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_ext_numba_type);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_ext_numba_type);
  __Pyx_GIVEREF(__pyx_v_ext_numba_type);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_6), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 286; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
  __pyx_t_6 = PyNumber_Add(__pyx_v_offsets, __pyx_t_2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 286; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_v_offsets);
  __pyx_v_offsets = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "numba/extension_types.pyx":287
 *     offsets = getattr(ext_type, '__numba_object_offset', [])
 *     offsets = offsets + compute_object_offsets(ext_numba_type, attrs_offset)
 *     ext_type.__numba_object_offset = offsets             # <<<<<<<<<<<<<<
 * 
 *     return ext_type
 */
  if (PyObject_SetAttr(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type), __pyx_n_s_1, __pyx_v_offsets) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 287; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":289
 *     ext_type.__numba_object_offset = offsets
 * 
 *     return ext_type             # <<<<<<<<<<<<<<
 * 
 * #------------------------------------------------------------------------
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_v_ext_type));
  __pyx_r = ((PyObject *)__pyx_cur_scope->__pyx_v_ext_type);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("numba.extension_types.create_new_extension_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_new);
  __Pyx_XDECREF(__pyx_v_struct_type);
  __Pyx_XDECREF(__pyx_v_struct_ctype);
  __Pyx_XDECREF(__pyx_v_upper);
  __Pyx_XDECREF(__pyx_v_vtab_p);
  __Pyx_XDECREF(__pyx_v_offsets);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5numba_15extension_types_9create_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5numba_15extension_types_9create_function = {__Pyx_NAMESTR("create_function"), (PyCFunction)__pyx_pw_5numba_15extension_types_9create_function, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_5numba_15extension_types_9create_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_methoddef = 0;
  PyObject *__pyx_v_py_func = 0;
  PyObject *__pyx_v_lfunc_pointer = 0;
  PyObject *__pyx_v_signature = 0;
  PyObject *__pyx_v_modname = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_function (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__methoddef,&__pyx_n_s__py_func,&__pyx_n_s__lfunc_pointer,&__pyx_n_s__signature,&__pyx_n_s__modname,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__methoddef)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__py_func)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_function", 1, 5, 5, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__lfunc_pointer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_function", 1, 5, 5, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__signature)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_function", 1, 5, 5, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__modname)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_function", 1, 5, 5, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_function") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_methoddef = values[0];
    __pyx_v_py_func = values[1];
    __pyx_v_lfunc_pointer = values[2];
    __pyx_v_signature = values[3];
    __pyx_v_modname = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_function", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("numba.extension_types.create_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5numba_15extension_types_8create_function(__pyx_self, __pyx_v_methoddef, __pyx_v_py_func, __pyx_v_lfunc_pointer, __pyx_v_signature, __pyx_v_modname);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numba/extension_types.pyx":295
 * #------------------------------------------------------------------------
 * 
 * def create_function(methoddef, py_func, lfunc_pointer, signature, modname):             # <<<<<<<<<<<<<<
 *     cdef Py_uintptr_t methoddef_p = ctypes.cast(ctypes.byref(methoddef),
 *                                                 ctypes.c_void_p).value
 */

static PyObject *__pyx_pf_5numba_15extension_types_8create_function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_methoddef, PyObject *__pyx_v_py_func, PyObject *__pyx_v_lfunc_pointer, PyObject *__pyx_v_signature, PyObject *__pyx_v_modname) {
  Py_uintptr_t __pyx_v_methoddef_p;
  PyMethodDef *__pyx_v_ml;
  Py_uintptr_t __pyx_v_lfunc_p;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_uintptr_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_function", 0);

  /* "numba/extension_types.pyx":296
 * 
 * def create_function(methoddef, py_func, lfunc_pointer, signature, modname):
 *     cdef Py_uintptr_t methoddef_p = ctypes.cast(ctypes.byref(methoddef),             # <<<<<<<<<<<<<<
 *                                                 ctypes.c_void_p).value
 *     cdef PyMethodDef *ml = <PyMethodDef *> methoddef_p
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__ctypes); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__cast); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__ctypes); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__byref); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_methoddef);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_methoddef);
  __Pyx_GIVEREF(__pyx_v_methoddef);
  __pyx_t_4 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":297
 * def create_function(methoddef, py_func, lfunc_pointer, signature, modname):
 *     cdef Py_uintptr_t methoddef_p = ctypes.cast(ctypes.byref(methoddef),
 *                                                 ctypes.c_void_p).value             # <<<<<<<<<<<<<<
 *     cdef PyMethodDef *ml = <PyMethodDef *> methoddef_p
 *     cdef Py_uintptr_t lfunc_p = lfunc_pointer
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__ctypes); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 297; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__c_void_p); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 297; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__value); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 297; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_from_py_Py_uintptr_t(__pyx_t_1); if (unlikely((__pyx_t_5 == (Py_uintptr_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 297; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_methoddef_p = __pyx_t_5;

  /* "numba/extension_types.pyx":298
 *     cdef Py_uintptr_t methoddef_p = ctypes.cast(ctypes.byref(methoddef),
 *                                                 ctypes.c_void_p).value
 *     cdef PyMethodDef *ml = <PyMethodDef *> methoddef_p             # <<<<<<<<<<<<<<
 *     cdef Py_uintptr_t lfunc_p = lfunc_pointer
 * 
 */
  __pyx_v_ml = ((PyMethodDef *)__pyx_v_methoddef_p);

  /* "numba/extension_types.pyx":299
 *                                                 ctypes.c_void_p).value
 *     cdef PyMethodDef *ml = <PyMethodDef *> methoddef_p
 *     cdef Py_uintptr_t lfunc_p = lfunc_pointer             # <<<<<<<<<<<<<<
 * 
 *     result = NumbaFunction_NewEx(ml, modname, getattr(py_func, "func_code", None),
 */
  __pyx_t_5 = __Pyx_PyInt_from_py_Py_uintptr_t(__pyx_v_lfunc_pointer); if (unlikely((__pyx_t_5 == (Py_uintptr_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 299; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_lfunc_p = __pyx_t_5;

  /* "numba/extension_types.pyx":301
 *     cdef Py_uintptr_t lfunc_p = lfunc_pointer
 * 
 *     result = NumbaFunction_NewEx(ml, modname, getattr(py_func, "func_code", None),             # <<<<<<<<<<<<<<
 *                                  NULL, <void *> lfunc_p, signature, py_func)
 *     return result
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_py_func, ((PyObject *)__pyx_n_s__func_code), Py_None); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 301; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);

  /* "numba/extension_types.pyx":302
 * 
 *     result = NumbaFunction_NewEx(ml, modname, getattr(py_func, "func_code", None),
 *                                  NULL, <void *> lfunc_p, signature, py_func)             # <<<<<<<<<<<<<<
 *     return result
 */
  __pyx_t_3 = NumbaFunction_NewEx(__pyx_v_ml, __pyx_v_modname, __pyx_t_1, NULL, ((void *)__pyx_v_lfunc_p), __pyx_v_signature, __pyx_v_py_func); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 301; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "numba/extension_types.pyx":303
 *     result = NumbaFunction_NewEx(ml, modname, getattr(py_func, "func_code", None),
 *                                  NULL, <void *> lfunc_p, signature, py_func)
 *     return result             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numba.extension_types.create_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_5numba_15extension_types___pyx_scope_struct__create_new_extension_type(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *p;
  PyObject *o;
  o = (*t->tp_alloc)(t, 0);
  if (!o) return 0;
  p = ((struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *)o);
  p->__pyx_v_ext_type = 0;
  p->__pyx_v_orig_new = 0;
  p->__pyx_v_vtab = 0;
  return o;
}

static void __pyx_tp_dealloc_5numba_15extension_types___pyx_scope_struct__create_new_extension_type(PyObject *o) {
  struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *p = (struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_ext_type);
  Py_CLEAR(p->__pyx_v_orig_new);
  Py_CLEAR(p->__pyx_v_vtab);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_5numba_15extension_types___pyx_scope_struct__create_new_extension_type(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *p = (struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *)o;
  if (p->__pyx_v_ext_type) {
    e = (*v)(p->__pyx_v_ext_type, a); if (e) return e;
  }
  if (p->__pyx_v_orig_new) {
    e = (*v)(p->__pyx_v_orig_new, a); if (e) return e;
  }
  if (p->__pyx_v_vtab) {
    e = (*v)(p->__pyx_v_vtab, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5numba_15extension_types___pyx_scope_struct__create_new_extension_type(PyObject *o) {
  struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *p = (struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type *)o;
  PyObject* tmp;
  tmp = ((PyObject*)p->__pyx_v_ext_type);
  p->__pyx_v_ext_type = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_orig_new);
  p->__pyx_v_orig_new = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_vtab);
  p->__pyx_v_vtab = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_5numba_15extension_types___pyx_scope_struct__create_new_extension_type[] = {
  {0, 0, 0, 0}
};

static PyNumberMethods __pyx_tp_as_number___pyx_scope_struct__create_new_extension_type = {
  0, /*nb_add*/
  0, /*nb_subtract*/
  0, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  0, /*nb_nonzero*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*nb_hex*/
  #endif
  0, /*nb_inplace_add*/
  0, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  #if PY_VERSION_HEX >= 0x02050000
  0, /*nb_index*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence___pyx_scope_struct__create_new_extension_type = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping___pyx_scope_struct__create_new_extension_type = {
  0, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer___pyx_scope_struct__create_new_extension_type = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  #if PY_VERSION_HEX >= 0x02060000
  0, /*bf_getbuffer*/
  #endif
  #if PY_VERSION_HEX >= 0x02060000
  0, /*bf_releasebuffer*/
  #endif
};

static PyTypeObject __pyx_type_5numba_15extension_types___pyx_scope_struct__create_new_extension_type = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("numba.extension_types.__pyx_scope_struct__create_new_extension_type"), /*tp_name*/
  sizeof(struct __pyx_obj_5numba_15extension_types___pyx_scope_struct__create_new_extension_type), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5numba_15extension_types___pyx_scope_struct__create_new_extension_type, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  &__pyx_tp_as_number___pyx_scope_struct__create_new_extension_type, /*tp_as_number*/
  &__pyx_tp_as_sequence___pyx_scope_struct__create_new_extension_type, /*tp_as_sequence*/
  &__pyx_tp_as_mapping___pyx_scope_struct__create_new_extension_type, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer___pyx_scope_struct__create_new_extension_type, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5numba_15extension_types___pyx_scope_struct__create_new_extension_type, /*tp_traverse*/
  __pyx_tp_clear_5numba_15extension_types___pyx_scope_struct__create_new_extension_type, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5numba_15extension_types___pyx_scope_struct__create_new_extension_type, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5numba_15extension_types___pyx_scope_struct__create_new_extension_type, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    __Pyx_NAMESTR("extension_types"),
    __Pyx_DOCSTR(__pyx_k_14), /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_1, __pyx_k_1, sizeof(__pyx_k_1), 0, 0, 1, 1},
  {&__pyx_kp_s_10, __pyx_k_10, sizeof(__pyx_k_10), 0, 0, 1, 0},
  {&__pyx_n_s_12, __pyx_k_12, sizeof(__pyx_k_12), 0, 0, 1, 1},
  {&__pyx_n_s_13, __pyx_k_13, sizeof(__pyx_k_13), 0, 0, 1, 1},
  {&__pyx_n_s_15, __pyx_k_15, sizeof(__pyx_k_15), 0, 0, 1, 1},
  {&__pyx_n_s_16, __pyx_k_16, sizeof(__pyx_k_16), 0, 0, 1, 1},
  {&__pyx_n_s_2, __pyx_k_2, sizeof(__pyx_k_2), 0, 0, 1, 1},
  {&__pyx_n_s_25, __pyx_k_25, sizeof(__pyx_k_25), 0, 0, 1, 1},
  {&__pyx_kp_s_6, __pyx_k_6, sizeof(__pyx_k_6), 0, 0, 1, 0},
  {&__pyx_n_s_7, __pyx_k_7, sizeof(__pyx_k_7), 0, 0, 1, 1},
  {&__pyx_n_s_8, __pyx_k_8, sizeof(__pyx_k_8), 0, 0, 1, 1},
  {&__pyx_n_s_9, __pyx_k_9, sizeof(__pyx_k_9), 0, 0, 1, 1},
  {&__pyx_n_s__NotImplementedError, __pyx_k__NotImplementedError, sizeof(__pyx_k__NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s____main__, __pyx_k____main__, sizeof(__pyx_k____main__), 0, 0, 1, 1},
  {&__pyx_n_s____new__, __pyx_k____new__, sizeof(__pyx_k____new__), 0, 0, 1, 1},
  {&__pyx_n_s____numba_attr_offset, __pyx_k____numba_attr_offset, sizeof(__pyx_k____numba_attr_offset), 0, 0, 1, 1},
  {&__pyx_n_s____numba_ext_type, __pyx_k____numba_ext_type, sizeof(__pyx_k____numba_ext_type), 0, 0, 1, 1},
  {&__pyx_n_s____numba_lfuncs, __pyx_k____numba_lfuncs, sizeof(__pyx_k____numba_lfuncs), 0, 0, 1, 1},
  {&__pyx_n_s____numba_obj_end, __pyx_k____numba_obj_end, sizeof(__pyx_k____numba_obj_end), 0, 0, 1, 1},
  {&__pyx_n_s____numba_orig_tp_new, __pyx_k____numba_orig_tp_new, sizeof(__pyx_k____numba_orig_tp_new), 0, 0, 1, 1},
  {&__pyx_n_s____numba_struct_type, __pyx_k____numba_struct_type, sizeof(__pyx_k____numba_struct_type), 0, 0, 1, 1},
  {&__pyx_n_s____numba_vtab, __pyx_k____numba_vtab, sizeof(__pyx_k____numba_vtab), 0, 0, 1, 1},
  {&__pyx_n_s____numba_vtab_offset, __pyx_k____numba_vtab_offset, sizeof(__pyx_k____numba_vtab_offset), 0, 0, 1, 1},
  {&__pyx_n_s____numba_vtab_p, __pyx_k____numba_vtab_p, sizeof(__pyx_k____numba_vtab_p), 0, 0, 1, 1},
  {&__pyx_n_s____numba_vtab_type, __pyx_k____numba_vtab_type, sizeof(__pyx_k____numba_vtab_type), 0, 0, 1, 1},
  {&__pyx_n_s____test__, __pyx_k____test__, sizeof(__pyx_k____test__), 0, 0, 1, 1},
  {&__pyx_n_s___numba_attrs, __pyx_k___numba_attrs, sizeof(__pyx_k___numba_attrs), 0, 0, 1, 1},
  {&__pyx_n_s__args, __pyx_k__args, sizeof(__pyx_k__args), 0, 0, 1, 1},
  {&__pyx_n_s__attribute_struct, __pyx_k__attribute_struct, sizeof(__pyx_k__attribute_struct), 0, 0, 1, 1},
  {&__pyx_n_s__attrs_offset, __pyx_k__attrs_offset, sizeof(__pyx_k__attrs_offset), 0, 0, 1, 1},
  {&__pyx_n_s__attrs_pointer, __pyx_k__attrs_pointer, sizeof(__pyx_k__attrs_pointer), 0, 0, 1, 1},
  {&__pyx_n_s__base_is_object, __pyx_k__base_is_object, sizeof(__pyx_k__base_is_object), 0, 0, 1, 1},
  {&__pyx_n_s__bases, __pyx_k__bases, sizeof(__pyx_k__bases), 0, 0, 1, 1},
  {&__pyx_n_s__byref, __pyx_k__byref, sizeof(__pyx_k__byref), 0, 0, 1, 1},
  {&__pyx_n_s__c_void_p, __pyx_k__c_void_p, sizeof(__pyx_k__c_void_p), 0, 0, 1, 1},
  {&__pyx_n_s__cast, __pyx_k__cast, sizeof(__pyx_k__cast), 0, 0, 1, 1},
  {&__pyx_n_s__cls, __pyx_k__cls, sizeof(__pyx_k__cls), 0, 0, 1, 1},
  {&__pyx_n_s__compute_vtab_offset, __pyx_k__compute_vtab_offset, sizeof(__pyx_k__compute_vtab_offset), 0, 0, 1, 1},
  {&__pyx_n_s__create_function, __pyx_k__create_function, sizeof(__pyx_k__create_function), 0, 0, 1, 1},
  {&__pyx_n_s__ctypes, __pyx_k__ctypes, sizeof(__pyx_k__ctypes), 0, 0, 1, 1},
  {&__pyx_n_s__dict, __pyx_k__dict, sizeof(__pyx_k__dict), 0, 0, 1, 1},
  {&__pyx_n_s__ext_numba_type, __pyx_k__ext_numba_type, sizeof(__pyx_k__ext_numba_type), 0, 0, 1, 1},
  {&__pyx_n_s__ext_type, __pyx_k__ext_type, sizeof(__pyx_k__ext_type), 0, 0, 1, 1},
  {&__pyx_n_s__ext_type_p, __pyx_k__ext_type_p, sizeof(__pyx_k__ext_type_p), 0, 0, 1, 1},
  {&__pyx_n_s__exttype, __pyx_k__exttype, sizeof(__pyx_k__exttype), 0, 0, 1, 1},
  {&__pyx_n_s__field_name, __pyx_k__field_name, sizeof(__pyx_k__field_name), 0, 0, 1, 1},
  {&__pyx_n_s__field_type, __pyx_k__field_type, sizeof(__pyx_k__field_type), 0, 0, 1, 1},
  {&__pyx_n_s__fields, __pyx_k__fields, sizeof(__pyx_k__fields), 0, 0, 1, 1},
  {&__pyx_n_s__func_code, __pyx_k__func_code, sizeof(__pyx_k__func_code), 0, 0, 1, 1},
  {&__pyx_n_s__get, __pyx_k__get, sizeof(__pyx_k__get), 0, 0, 1, 1},
  {&__pyx_n_s__is_array, __pyx_k__is_array, sizeof(__pyx_k__is_array), 0, 0, 1, 1},
  {&__pyx_n_s__is_object, __pyx_k__is_object, sizeof(__pyx_k__is_object), 0, 0, 1, 1},
  {&__pyx_n_s__kwds, __pyx_k__kwds, sizeof(__pyx_k__kwds), 0, 0, 1, 1},
  {&__pyx_n_s__lfunc_p, __pyx_k__lfunc_p, sizeof(__pyx_k__lfunc_p), 0, 0, 1, 1},
  {&__pyx_n_s__lfunc_pointer, __pyx_k__lfunc_pointer, sizeof(__pyx_k__lfunc_pointer), 0, 0, 1, 1},
  {&__pyx_n_s__llvm_methods, __pyx_k__llvm_methods, sizeof(__pyx_k__llvm_methods), 0, 0, 1, 1},
  {&__pyx_n_s__method_pointers, __pyx_k__method_pointers, sizeof(__pyx_k__method_pointers), 0, 0, 1, 1},
  {&__pyx_n_s__methoddef, __pyx_k__methoddef, sizeof(__pyx_k__methoddef), 0, 0, 1, 1},
  {&__pyx_n_s__methoddef_p, __pyx_k__methoddef_p, sizeof(__pyx_k__methoddef_p), 0, 0, 1, 1},
  {&__pyx_n_s__ml, __pyx_k__ml, sizeof(__pyx_k__ml), 0, 0, 1, 1},
  {&__pyx_n_s__modname, __pyx_k__modname, sizeof(__pyx_k__modname), 0, 0, 1, 1},
  {&__pyx_n_s__name, __pyx_k__name, sizeof(__pyx_k__name), 0, 0, 1, 1},
  {&__pyx_n_s__new, __pyx_k__new, sizeof(__pyx_k__new), 0, 0, 1, 1},
  {&__pyx_n_s__new_func, __pyx_k__new_func, sizeof(__pyx_k__new_func), 0, 0, 1, 1},
  {&__pyx_n_s__numba, __pyx_k__numba, sizeof(__pyx_k__numba), 0, 0, 1, 1},
  {&__pyx_n_s__obj, __pyx_k__obj, sizeof(__pyx_k__obj), 0, 0, 1, 1},
  {&__pyx_n_s__obj_p, __pyx_k__obj_p, sizeof(__pyx_k__obj_p), 0, 0, 1, 1},
  {&__pyx_n_s__object, __pyx_k__object, sizeof(__pyx_k__object), 0, 0, 1, 1},
  {&__pyx_n_s__offset, __pyx_k__offset, sizeof(__pyx_k__offset), 0, 0, 1, 1},
  {&__pyx_n_s__offsetof, __pyx_k__offsetof, sizeof(__pyx_k__offsetof), 0, 0, 1, 1},
  {&__pyx_n_s__offsets, __pyx_k__offsets, sizeof(__pyx_k__offsets), 0, 0, 1, 1},
  {&__pyx_n_s__orig_new, __pyx_k__orig_new, sizeof(__pyx_k__orig_new), 0, 0, 1, 1},
  {&__pyx_n_s__pointer, __pyx_k__pointer, sizeof(__pyx_k__pointer), 0, 0, 1, 1},
  {&__pyx_n_s__py_class, __pyx_k__py_class, sizeof(__pyx_k__py_class), 0, 0, 1, 1},
  {&__pyx_n_s__py_func, __pyx_k__py_func, sizeof(__pyx_k__py_func), 0, 0, 1, 1},
  {&__pyx_n_s__result, __pyx_k__result, sizeof(__pyx_k__result), 0, 0, 1, 1},
  {&__pyx_n_s__signature, __pyx_k__signature, sizeof(__pyx_k__signature), 0, 0, 1, 1},
  {&__pyx_n_s__sizeof, __pyx_k__sizeof, sizeof(__pyx_k__sizeof), 0, 0, 1, 1},
  {&__pyx_n_s__staticmethod, __pyx_k__staticmethod, sizeof(__pyx_k__staticmethod), 0, 0, 1, 1},
  {&__pyx_n_s__struct_ctype, __pyx_k__struct_ctype, sizeof(__pyx_k__struct_ctype), 0, 0, 1, 1},
  {&__pyx_n_s__struct_offset, __pyx_k__struct_offset, sizeof(__pyx_k__struct_offset), 0, 0, 1, 1},
  {&__pyx_n_s__struct_type, __pyx_k__struct_type, sizeof(__pyx_k__struct_type), 0, 0, 1, 1},
  {&__pyx_n_s__super, __pyx_k__super, sizeof(__pyx_k__super), 0, 0, 1, 1},
  {&__pyx_n_s__sys, __pyx_k__sys, sizeof(__pyx_k__sys), 0, 0, 1, 1},
  {&__pyx_n_s__to_ctypes, __pyx_k__to_ctypes, sizeof(__pyx_k__to_ctypes), 0, 0, 1, 1},
  {&__pyx_n_s__type_p, __pyx_k__type_p, sizeof(__pyx_k__type_p), 0, 0, 1, 1},
  {&__pyx_n_s__upper, __pyx_k__upper, sizeof(__pyx_k__upper), 0, 0, 1, 1},
  {&__pyx_n_s__value, __pyx_k__value, sizeof(__pyx_k__value), 0, 0, 1, 1},
  {&__pyx_n_s__vtab, __pyx_k__vtab, sizeof(__pyx_k__vtab), 0, 0, 1, 1},
  {&__pyx_n_s__vtab_location, __pyx_k__vtab_location, sizeof(__pyx_k__vtab_location), 0, 0, 1, 1},
  {&__pyx_n_s__vtab_offset, __pyx_k__vtab_offset, sizeof(__pyx_k__vtab_offset), 0, 0, 1, 1},
  {&__pyx_n_s__vtab_p, __pyx_k__vtab_p, sizeof(__pyx_k__vtab_p), 0, 0, 1, 1},
  {&__pyx_n_s__vtab_type, __pyx_k__vtab_type, sizeof(__pyx_k__vtab_type), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_staticmethod = __Pyx_GetName(__pyx_b, __pyx_n_s__staticmethod); if (!__pyx_builtin_staticmethod) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 234; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_object = __Pyx_GetName(__pyx_b, __pyx_n_s__object); if (!__pyx_builtin_object) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_NotImplementedError = __Pyx_GetName(__pyx_b, __pyx_n_s__NotImplementedError); if (!__pyx_builtin_NotImplementedError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 255; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_super = __Pyx_GetName(__pyx_b, __pyx_n_s__super); if (!__pyx_builtin_super) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 204; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "numba/extension_types.pyx":194
 *     cdef Py_ssize_t vtab_offset, attrs_offset
 * 
 *     orig_new = dict.get('__new__', None)             # <<<<<<<<<<<<<<
 *     def new(cls, *args, **kwds):
 *         "Create a new object and patch it with a vtab"
 */
  __pyx_k_tuple_3 = PyTuple_Pack(2, ((PyObject *)__pyx_n_s____new__), Py_None); if (unlikely(!__pyx_k_tuple_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_3);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_3));

  /* "numba/extension_types.pyx":195
 * 
 *     orig_new = dict.get('__new__', None)
 *     def new(cls, *args, **kwds):             # <<<<<<<<<<<<<<
 *         "Create a new object and patch it with a vtab"
 *         cdef PyObject *obj_p
 */
  __pyx_k_tuple_4 = PyTuple_Pack(10, ((PyObject *)__pyx_n_s__cls), ((PyObject *)__pyx_n_s__args), ((PyObject *)__pyx_n_s__kwds), ((PyObject *)__pyx_n_s__args), ((PyObject *)__pyx_n_s__kwds), ((PyObject *)__pyx_n_s__obj_p), ((PyObject *)__pyx_n_s__vtab_location), ((PyObject *)__pyx_n_s__new_func), ((PyObject *)__pyx_n_s__obj), ((PyObject *)__pyx_n_s__attrs_pointer)); if (unlikely(!__pyx_k_tuple_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_4);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_4));
  __pyx_k_codeobj_5 = (PyObject*)__Pyx_PyCode_New(1, 0, 10, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_4, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_6, __pyx_n_s__new, 195, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":255
 *     ext_type_p.tp_basicsize = upper
 *     if ext_type_p.tp_itemsize:
 *         raise NotImplementedError("Subclassing variable sized objects")             # <<<<<<<<<<<<<<
 * 
 *     # TODO: Figure out ctypes policy on tp_dealloc/tp_traverse/tp_clear
 */
  __pyx_k_tuple_11 = PyTuple_Pack(1, ((PyObject *)__pyx_kp_s_10)); if (unlikely(!__pyx_k_tuple_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 255; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_11);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_11));

  /* "numba/extension_types.pyx":62
 *     return <void *> align(<Py_uintptr_t> memory, alignment)
 * 
 * def compute_vtab_offset(py_class):             # <<<<<<<<<<<<<<
 *     "Returns the vtab pointer offset in the object"
 *     offset = getattr(py_class, '__numba_vtab_offset', None)
 */
  __pyx_k_tuple_17 = PyTuple_Pack(3, ((PyObject *)__pyx_n_s__py_class), ((PyObject *)__pyx_n_s__offset), ((PyObject *)__pyx_n_s__type_p)); if (unlikely(!__pyx_k_tuple_17)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_17);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_17));
  __pyx_k_codeobj_18 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_6, __pyx_n_s__compute_vtab_offset, 62, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_18)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":71
 *     return align(type_p.tp_basicsize, 8)
 * 
 * def compute_attrs_offset(py_class):             # <<<<<<<<<<<<<<
 *     "Returns the start of the attribute struct"
 *     offset = getattr(py_class, '__numba_attr_offset', None)
 */
  __pyx_k_tuple_19 = PyTuple_Pack(2, ((PyObject *)__pyx_n_s__py_class), ((PyObject *)__pyx_n_s__offset)); if (unlikely(!__pyx_k_tuple_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 71; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_19);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_19));
  __pyx_k_codeobj_20 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_19, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_6, __pyx_n_s_9, 71, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 71; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":88
 * # on the extension type object.
 * 
 * def compute_object_offsets(ext_type, struct_offset):             # <<<<<<<<<<<<<<
 *     """
 *     Given an typesystem.ExtensionType, compute the offsets of object
 */
  __pyx_k_tuple_21 = PyTuple_Pack(6, ((PyObject *)__pyx_n_s__ext_type), ((PyObject *)__pyx_n_s__struct_offset), ((PyObject *)__pyx_n_s__offsets), ((PyObject *)__pyx_n_s__field_name), ((PyObject *)__pyx_n_s__field_type), ((PyObject *)__pyx_n_s__offset)); if (unlikely(!__pyx_k_tuple_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 88; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_21);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_21));
  __pyx_k_codeobj_22 = (PyObject*)__Pyx_PyCode_New(2, 0, 6, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_6, __pyx_n_s_13, 88, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 88; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":184
 * #------------------------------------------------------------------------
 * 
 * def create_new_extension_type(name, bases, dict, ext_numba_type,             # <<<<<<<<<<<<<<
 *                               vtab, vtab_type, llvm_methods, method_pointers):
 *     """
 */
  __pyx_k_tuple_23 = PyTuple_Pack(21, ((PyObject *)__pyx_n_s__name), ((PyObject *)__pyx_n_s__bases), ((PyObject *)__pyx_n_s__dict), ((PyObject *)__pyx_n_s__ext_numba_type), ((PyObject *)__pyx_n_s__vtab), ((PyObject *)__pyx_n_s__vtab_type), ((PyObject *)__pyx_n_s__llvm_methods), ((PyObject *)__pyx_n_s__method_pointers), ((PyObject *)__pyx_n_s__ext_type_p), ((PyObject *)__pyx_n_s__vtab_offset), ((PyObject *)__pyx_n_s__attrs_offset), ((PyObject *)__pyx_n_s__orig_new), ((PyObject *)__pyx_n_s__new), ((PyObject *)__pyx_n_s__new), ((PyObject *)__pyx_n_s__ext_type), ((PyObject *)__pyx_n_s__base_is_object), ((PyObject *)__pyx_n_s__struct_type), ((PyObject *)__pyx_n_s__struct_ctype), ((PyObject *)__pyx_n_s__upper), ((PyObject *)__pyx_n_s__vtab_p), ((PyObject *)__pyx_n_s__offsets)); if (unlikely(!__pyx_k_tuple_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_23);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_23));
  __pyx_k_codeobj_24 = (PyObject*)__Pyx_PyCode_New(8, 0, 21, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_6, __pyx_n_s_25, 184, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":295
 * #------------------------------------------------------------------------
 * 
 * def create_function(methoddef, py_func, lfunc_pointer, signature, modname):             # <<<<<<<<<<<<<<
 *     cdef Py_uintptr_t methoddef_p = ctypes.cast(ctypes.byref(methoddef),
 *                                                 ctypes.c_void_p).value
 */
  __pyx_k_tuple_26 = PyTuple_Pack(9, ((PyObject *)__pyx_n_s__methoddef), ((PyObject *)__pyx_n_s__py_func), ((PyObject *)__pyx_n_s__lfunc_pointer), ((PyObject *)__pyx_n_s__signature), ((PyObject *)__pyx_n_s__modname), ((PyObject *)__pyx_n_s__methoddef_p), ((PyObject *)__pyx_n_s__ml), ((PyObject *)__pyx_n_s__lfunc_p), ((PyObject *)__pyx_n_s__result)); if (unlikely(!__pyx_k_tuple_26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_26);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_26));
  __pyx_k_codeobj_27 = (PyObject*)__Pyx_PyCode_New(5, 0, 9, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_6, __pyx_n_s__create_function, 295, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_27)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initextension_types(void); /*proto*/
PyMODINIT_FUNC initextension_types(void)
#else
PyMODINIT_FUNC PyInit_extension_types(void); /*proto*/
PyMODINIT_FUNC PyInit_extension_types(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_extension_types(void)", 0);
  if ( __Pyx_check_binary_version() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #ifdef __Pyx_CyFunction_USED
  if (__Pyx_CyFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4(__Pyx_NAMESTR("extension_types"), __pyx_methods, __Pyx_DOCSTR(__pyx_k_14), 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (!PyDict_GetItemString(modules, "numba.extension_types")) {
      if (unlikely(PyDict_SetItemString(modules, "numba.extension_types", __pyx_m) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }
  #endif
  __pyx_b = PyImport_AddModule(__Pyx_NAMESTR(__Pyx_BUILTIN_MODULE_NAME)); if (unlikely(!__pyx_b)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (__Pyx_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  /*--- Initialize various global constants etc. ---*/
  if (unlikely(__Pyx_InitGlobals() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_module_is_main_numba__extension_types) {
    if (__Pyx_SetAttrString(__pyx_m, "__name__", __pyx_n_s____main__) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  }
  /*--- Builtin init code ---*/
  if (unlikely(__Pyx_InitCachedBuiltins() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Constants init code ---*/
  if (unlikely(__Pyx_InitCachedConstants() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_5numba_15extension_types___pyx_scope_struct__create_new_extension_type) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numba_15extension_types___pyx_scope_struct__create_new_extension_type = &__pyx_type_5numba_15extension_types___pyx_scope_struct__create_new_extension_type;
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 9; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), 0); if (unlikely(!__pyx_ptype_7cpython_4bool_bool)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 8; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), 0); if (unlikely(!__pyx_ptype_7cpython_7complex_complex)) {__pyx_filename = __pyx_f[3]; __pyx_lineno = 15; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/

  /* "numba/extension_types.pyx":10
 * from numba._numba cimport *
 * 
 * import sys             # <<<<<<<<<<<<<<
 * import ctypes
 * 
 */
  __pyx_t_1 = __Pyx_Import(((PyObject *)__pyx_n_s__sys), 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 10; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__sys, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 10; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":11
 * 
 * import sys
 * import ctypes             # <<<<<<<<<<<<<<
 * 
 * import numba
 */
  __pyx_t_1 = __Pyx_Import(((PyObject *)__pyx_n_s__ctypes), 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 11; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__ctypes, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 11; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":13
 * import ctypes
 * 
 * import numba             # <<<<<<<<<<<<<<
 * 
 * ctypedef object (*tp_new_func)(PyObject *, PyObject *, PyObject *)
 */
  __pyx_t_1 = __Pyx_Import(((PyObject *)__pyx_n_s__numba), 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 13; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__numba, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 13; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":41
 *                     void *native_func, native_signature, keep_alive)
 * 
 * NumbaFunction_init()             # <<<<<<<<<<<<<<
 * NumbaFunction_NewEx_pointer = <Py_uintptr_t> &NumbaFunction_NewEx
 * 
 */
  __pyx_t_2 = NumbaFunction_init(); if (unlikely(__pyx_t_2 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 41; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "numba/extension_types.pyx":42
 * 
 * NumbaFunction_init()
 * NumbaFunction_NewEx_pointer = <Py_uintptr_t> &NumbaFunction_NewEx             # <<<<<<<<<<<<<<
 * 
 * numbafunc_closure_field_offset = closure_field_offset
 */
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_uintptr_t(((Py_uintptr_t)(&NumbaFunction_NewEx))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 42; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_15, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 42; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":44
 * NumbaFunction_NewEx_pointer = <Py_uintptr_t> &NumbaFunction_NewEx
 * 
 * numbafunc_closure_field_offset = closure_field_offset             # <<<<<<<<<<<<<<
 * 
 * cdef Py_uintptr_t align(Py_uintptr_t p, size_t alignment) nogil:
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(closure_field_offset); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 44; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_16, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 44; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":62
 *     return <void *> align(<Py_uintptr_t> memory, alignment)
 * 
 * def compute_vtab_offset(py_class):             # <<<<<<<<<<<<<<
 *     "Returns the vtab pointer offset in the object"
 *     offset = getattr(py_class, '__numba_vtab_offset', None)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5numba_15extension_types_1compute_vtab_offset, NULL, __pyx_n_s_8); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__compute_vtab_offset, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 62; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":71
 *     return align(type_p.tp_basicsize, 8)
 * 
 * def compute_attrs_offset(py_class):             # <<<<<<<<<<<<<<
 *     "Returns the start of the attribute struct"
 *     offset = getattr(py_class, '__numba_attr_offset', None)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5numba_15extension_types_3compute_attrs_offset, NULL, __pyx_n_s_8); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 71; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_9, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 71; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":88
 * # on the extension type object.
 * 
 * def compute_object_offsets(ext_type, struct_offset):             # <<<<<<<<<<<<<<
 *     """
 *     Given an typesystem.ExtensionType, compute the offsets of object
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5numba_15extension_types_5compute_object_offsets, NULL, __pyx_n_s_8); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 88; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_13, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 88; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":184
 * #------------------------------------------------------------------------
 * 
 * def create_new_extension_type(name, bases, dict, ext_numba_type,             # <<<<<<<<<<<<<<
 *                               vtab, vtab_type, llvm_methods, method_pointers):
 *     """
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5numba_15extension_types_7create_new_extension_type, NULL, __pyx_n_s_8); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_25, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":295
 * #------------------------------------------------------------------------
 * 
 * def create_function(methoddef, py_func, lfunc_pointer, signature, modname):             # <<<<<<<<<<<<<<
 *     cdef Py_uintptr_t methoddef_p = ctypes.cast(ctypes.byref(methoddef),
 *                                                 ctypes.c_void_p).value
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5numba_15extension_types_9create_function, NULL, __pyx_n_s_8); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__create_function, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numba/extension_types.pyx":1
 * """             # <<<<<<<<<<<<<<
 * This modules creates extension types.
 * 
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s____test__, ((PyObject *)__pyx_t_1)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    __Pyx_AddTraceback("init numba.extension_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init numba.extension_types");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* Runtime support code */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif /* CYTHON_REFNANNY */

static PyObject *__Pyx_GetName(PyObject *dict, PyObject *name) {
    PyObject *result;
    result = PyObject_GetAttr(dict, name);
    if (!result) {
        if (dict != __pyx_b) {
            PyErr_Clear();
            result = PyObject_GetAttr(__pyx_b, name);
        }
        if (!result) {
            PyErr_SetObject(PyExc_NameError, name);
        }
    }
    return result;
}

static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = PyObject_GetAttr(o, n);
    if (!r) {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError))
            goto bad;
        PyErr_Clear();
        r = d;
        Py_INCREF(d);
    }
    return r;
bad:
    return NULL;
}

static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%s() takes %s %" CYTHON_FORMAT_SSIZE_T "d positional argument%s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%s() got an unexpected keyword argument '%s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%s to unpack",
                 index, (index == 1) ? "" : "s");
}

static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL)
        value = Py_None; /* Mark as deleted */
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    PyObject* dict = PyModule_GetDict(__pyx_m);
    Py_XINCREF(dict);
    return dict;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op)
{
    if (op->defaults_tuple) {
        Py_INCREF(op->defaults_tuple);
        return op->defaults_tuple;
    }
    if (op->defaults_getter) {
        PyObject *res = op->defaults_getter((PyObject *) op);
        if (likely(res)) {
            Py_INCREF(res);
            op->defaults_tuple = res;
        }
        return res;
    }
    Py_INCREF(Py_None);
    return Py_None;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, 0, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, 0, 0, 0},
    {0, 0, 0, 0, 0}
};
#ifndef PY_WRITE_RESTRICTED /* < Py2.5 */
#define PY_WRITE_RESTRICTED WRITE_RESTRICTED
#endif
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {__Pyx_NAMESTR("__reduce__"), (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    op->func_weakreflist = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_getter = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyMem_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (m->func_weakreflist != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return PyMethod_New(func,
                            type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
#if CYTHON_COMPILING_IN_PYPY
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    Py_ssize_t size;
    switch (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)) {
    case METH_VARARGS:
        if (likely(kw == NULL) || PyDict_Size(kw) == 0)
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL) || PyDict_Size(kw) == 0) {
            size = PyTuple_GET_SIZE(arg);
            if (size == 0)
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%zd given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL) || PyDict_Size(kw) == 0) {
            size = PyTuple_GET_SIZE(arg);
            if (size == 1)
                return (*meth)(self, PyTuple_GET_ITEM(arg, 0));
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%zd given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
#else
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
	return PyCFunction_Call(func, arg, kw);
}
#endif
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __Pyx_NAMESTR("cython_function_or_method"), /*tp_name*/
    sizeof(__pyx_CyFunctionObject),   /*tp_basicsize*/
    0,                                  /*tp_itemsize*/
    (destructor) __Pyx_CyFunction_dealloc, /*tp_dealloc*/
    0,                                  /*tp_print*/
    0,                                  /*tp_getattr*/
    0,                                  /*tp_setattr*/
#if PY_MAJOR_VERSION < 3
    0,                                  /*tp_compare*/
#else
    0,                                  /*reserved*/
#endif
    (reprfunc) __Pyx_CyFunction_repr,   /*tp_repr*/
    0,                                  /*tp_as_number*/
    0,                                  /*tp_as_sequence*/
    0,                                  /*tp_as_mapping*/
    0,                                  /*tp_hash*/
    __Pyx_CyFunction_Call,              /*tp_call*/
    0,                                  /*tp_str*/
    0,                                  /*tp_getattro*/
    0,                                  /*tp_setattro*/
    0,                                  /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC, /* tp_flags*/
    0,                                  /*tp_doc*/
    (traverseproc) __Pyx_CyFunction_traverse,   /*tp_traverse*/
    (inquiry) __Pyx_CyFunction_clear,   /*tp_clear*/
    0,                                  /*tp_richcompare*/
    offsetof(__pyx_CyFunctionObject, func_weakreflist), /* tp_weaklistoffse */
    0,                                  /*tp_iter*/
    0,                                  /*tp_iternext*/
    __pyx_CyFunction_methods,           /*tp_methods*/
    __pyx_CyFunction_members,           /*tp_members*/
    __pyx_CyFunction_getsets,           /*tp_getset*/
    0,                                  /*tp_base*/
    0,                                  /*tp_dict*/
    __Pyx_CyFunction_descr_get,         /*tp_descr_get*/
    0,                                  /*tp_descr_set*/
    offsetof(__pyx_CyFunctionObject, func_dict),/*tp_dictoffset*/
    0,                                  /*tp_init*/
    0,                                  /*tp_alloc*/
    0,                                  /*tp_new*/
    0,                                  /*tp_free*/
    0,                                  /*tp_is_gc*/
    0,                                  /*tp_bases*/
    0,                                  /*tp_mro*/
    0,                                  /*tp_cache*/
    0,                                  /*tp_subclasses*/
    0,                                  /*tp_weaklist*/
    0,                                  /*tp_del*/
#if PY_VERSION_HEX >= 0x02060000
    0,                                  /*tp_version_tag*/
#endif
};
static int __Pyx_CyFunction_init(void) {
#if !CYTHON_COMPILING_IN_PYPY
    __pyx_CyFunctionType_type.tp_call = PyCFunction_Call;
#endif
    if (PyType_Ready(&__pyx_CyFunctionType_type) < 0)
        return -1;
    __pyx_CyFunctionType = &__pyx_CyFunctionType_type;
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyMem_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_Restore(type, value, tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(type, value, tb);
#endif
}

#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    #if PY_VERSION_HEX < 0x02050000
    if (PyClass_Check(type)) {
    #else
    if (PyType_Check(type)) {
    #endif
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        #if PY_VERSION_HEX < 0x02050000
        if (PyInstance_Check(type)) {
            type = (PyObject*) ((PyInstanceObject*)type)->in_class;
            Py_INCREF(type);
        } else {
            type = 0;
            PyErr_SetString(PyExc_TypeError,
                "raise: exception must be an old-style class or instance");
            goto raise_error;
        }
        #else
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
        #endif
    }
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else /* Python 3+ */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *args;
        if (!value)
            args = PyTuple_New(0);
        else if (PyTuple_Check(value)) {
            Py_INCREF(value);
            args = value;
        } else
            args = PyTuple_Pack(1, value);
        if (!args)
            goto bad;
        owned_instance = PyEval_CallObject(type, args);
        Py_DECREF(args);
        if (!owned_instance)
            goto bad;
        value = owned_instance;
        if (!PyExceptionInstance_Check(value)) {
            PyErr_Format(PyExc_TypeError,
                         "calling %R should have returned an instance of "
                         "BaseException, not %R",
                         type, Py_TYPE(value));
            goto bad;
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import = 0;
    py_import = __Pyx_GetAttrString(__pyx_b, "__import__");
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    #if PY_VERSION_HEX >= 0x02050000
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0; /* try absolute import on failure */
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
    #else
    if (level>0) {
        PyErr_SetString(PyExc_RuntimeError, "Relative import is not supported for Python <=2.4.");
        goto bad;
    }
    module = PyObject_CallFunctionObjArgs(py_import,
        name, global_dict, empty_dict, list, NULL);
    #endif
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_Py_uintptr_t(Py_uintptr_t val) {
    const Py_uintptr_t neg_one = (Py_uintptr_t)-1, const_zero = (Py_uintptr_t)0;
    const int is_unsigned = const_zero < neg_one;
    if ((sizeof(Py_uintptr_t) == sizeof(char))  ||
        (sizeof(Py_uintptr_t) == sizeof(short))) {
        return PyInt_FromLong((long)val);
    } else if ((sizeof(Py_uintptr_t) == sizeof(int)) ||
               (sizeof(Py_uintptr_t) == sizeof(long))) {
        if (is_unsigned)
            return PyLong_FromUnsignedLong((unsigned long)val);
        else
            return PyInt_FromLong((long)val);
    } else if (sizeof(Py_uintptr_t) == sizeof(PY_LONG_LONG)) {
        if (is_unsigned)
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG)val);
        else
            return PyLong_FromLongLong((PY_LONG_LONG)val);
    } else {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        return _PyLong_FromByteArray(bytes, sizeof(Py_uintptr_t),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE Py_uintptr_t __Pyx_PyInt_from_py_Py_uintptr_t(PyObject* x) {
    const Py_uintptr_t neg_one = (Py_uintptr_t)-1, const_zero = (Py_uintptr_t)0;
    const int is_unsigned = const_zero < neg_one;
    if (sizeof(Py_uintptr_t) == sizeof(char)) {
        if (is_unsigned)
            return (Py_uintptr_t)__Pyx_PyInt_AsUnsignedChar(x);
        else
            return (Py_uintptr_t)__Pyx_PyInt_AsSignedChar(x);
    } else if (sizeof(Py_uintptr_t) == sizeof(short)) {
        if (is_unsigned)
            return (Py_uintptr_t)__Pyx_PyInt_AsUnsignedShort(x);
        else
            return (Py_uintptr_t)__Pyx_PyInt_AsSignedShort(x);
    } else if (sizeof(Py_uintptr_t) == sizeof(int)) {
        if (is_unsigned)
            return (Py_uintptr_t)__Pyx_PyInt_AsUnsignedInt(x);
        else
            return (Py_uintptr_t)__Pyx_PyInt_AsSignedInt(x);
    } else if (sizeof(Py_uintptr_t) == sizeof(long)) {
        if (is_unsigned)
            return (Py_uintptr_t)__Pyx_PyInt_AsUnsignedLong(x);
        else
            return (Py_uintptr_t)__Pyx_PyInt_AsSignedLong(x);
    } else if (sizeof(Py_uintptr_t) == sizeof(PY_LONG_LONG)) {
        if (is_unsigned)
            return (Py_uintptr_t)__Pyx_PyInt_AsUnsignedLongLong(x);
        else
            return (Py_uintptr_t)__Pyx_PyInt_AsSignedLongLong(x);
    }  else {
        #if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
        #else
        Py_uintptr_t val;
        PyObject *v = __Pyx_PyNumber_Int(x);
        #if PY_VERSION_HEX < 0x03000000
        if (likely(v) && !PyLong_Check(v)) {
            PyObject *tmp = v;
            v = PyNumber_Long(tmp);
            Py_DECREF(tmp);
        }
        #endif
        if (likely(v)) {
            int one = 1; int is_little = (int)*(unsigned char *)&one;
            unsigned char *bytes = (unsigned char *)&val;
            int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                          bytes, sizeof(val),
                                          is_little, !is_unsigned);
            Py_DECREF(v);
            if (likely(!ret))
                return val;
        }
        #endif
        return (Py_uintptr_t)-1;
    }
}

static CYTHON_INLINE unsigned char __Pyx_PyInt_AsUnsignedChar(PyObject* x) {
    const unsigned char neg_one = (unsigned char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned char" :
                    "value too large to convert to unsigned char");
            }
            return (unsigned char)-1;
        }
        return (unsigned char)val;
    }
    return (unsigned char)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE unsigned short __Pyx_PyInt_AsUnsignedShort(PyObject* x) {
    const unsigned short neg_one = (unsigned short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned short" :
                    "value too large to convert to unsigned short");
            }
            return (unsigned short)-1;
        }
        return (unsigned short)val;
    }
    return (unsigned short)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE unsigned int __Pyx_PyInt_AsUnsignedInt(PyObject* x) {
    const unsigned int neg_one = (unsigned int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned int" :
                    "value too large to convert to unsigned int");
            }
            return (unsigned int)-1;
        }
        return (unsigned int)val;
    }
    return (unsigned int)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE char __Pyx_PyInt_AsChar(PyObject* x) {
    const char neg_one = (char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to char" :
                    "value too large to convert to char");
            }
            return (char)-1;
        }
        return (char)val;
    }
    return (char)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE short __Pyx_PyInt_AsShort(PyObject* x) {
    const short neg_one = (short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to short" :
                    "value too large to convert to short");
            }
            return (short)-1;
        }
        return (short)val;
    }
    return (short)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE int __Pyx_PyInt_AsInt(PyObject* x) {
    const int neg_one = (int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to int" :
                    "value too large to convert to int");
            }
            return (int)-1;
        }
        return (int)val;
    }
    return (int)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE signed char __Pyx_PyInt_AsSignedChar(PyObject* x) {
    const signed char neg_one = (signed char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed char" :
                    "value too large to convert to signed char");
            }
            return (signed char)-1;
        }
        return (signed char)val;
    }
    return (signed char)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE signed short __Pyx_PyInt_AsSignedShort(PyObject* x) {
    const signed short neg_one = (signed short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed short" :
                    "value too large to convert to signed short");
            }
            return (signed short)-1;
        }
        return (signed short)val;
    }
    return (signed short)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE signed int __Pyx_PyInt_AsSignedInt(PyObject* x) {
    const signed int neg_one = (signed int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed int" :
                    "value too large to convert to signed int");
            }
            return (signed int)-1;
        }
        return (signed int)val;
    }
    return (signed int)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE int __Pyx_PyInt_AsLongDouble(PyObject* x) {
    const int neg_one = (int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to int" :
                    "value too large to convert to int");
            }
            return (int)-1;
        }
        return (int)val;
    }
    return (int)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE unsigned long __Pyx_PyInt_AsUnsignedLong(PyObject* x) {
    const unsigned long neg_one = (unsigned long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to unsigned long");
            return (unsigned long)-1;
        }
        return (unsigned long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to unsigned long");
                return (unsigned long)-1;
            }
            return (unsigned long)PyLong_AsUnsignedLong(x);
        } else {
            return (unsigned long)PyLong_AsLong(x);
        }
    } else {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned long)-1;
        val = __Pyx_PyInt_AsUnsignedLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_AsUnsignedLongLong(PyObject* x) {
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to unsigned PY_LONG_LONG");
            return (unsigned PY_LONG_LONG)-1;
        }
        return (unsigned PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to unsigned PY_LONG_LONG");
                return (unsigned PY_LONG_LONG)-1;
            }
            return (unsigned PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
            return (unsigned PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        unsigned PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsUnsignedLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE long __Pyx_PyInt_AsLong(PyObject* x) {
    const long neg_one = (long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to long");
            return (long)-1;
        }
        return (long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to long");
                return (long)-1;
            }
            return (long)PyLong_AsUnsignedLong(x);
        } else {
            return (long)PyLong_AsLong(x);
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (long)-1;
        val = __Pyx_PyInt_AsLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_AsLongLong(PyObject* x) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to PY_LONG_LONG");
            return (PY_LONG_LONG)-1;
        }
        return (PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to PY_LONG_LONG");
                return (PY_LONG_LONG)-1;
            }
            return (PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
            return (PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE signed long __Pyx_PyInt_AsSignedLong(PyObject* x) {
    const signed long neg_one = (signed long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to signed long");
            return (signed long)-1;
        }
        return (signed long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to signed long");
                return (signed long)-1;
            }
            return (signed long)PyLong_AsUnsignedLong(x);
        } else {
            return (signed long)PyLong_AsLong(x);
        }
    } else {
        signed long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (signed long)-1;
        val = __Pyx_PyInt_AsSignedLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE signed PY_LONG_LONG __Pyx_PyInt_AsSignedLongLong(PyObject* x) {
    const signed PY_LONG_LONG neg_one = (signed PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to signed PY_LONG_LONG");
            return (signed PY_LONG_LONG)-1;
        }
        return (signed PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to signed PY_LONG_LONG");
                return (signed PY_LONG_LONG)-1;
            }
            return (signed PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
            return (signed PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        signed PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (signed PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsSignedLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
}

static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        #if PY_VERSION_HEX < 0x02050000
        return PyErr_Warn(NULL, message);
        #else
        return PyErr_WarnEx(NULL, message, 1);
        #endif
    }
    return 0;
}

#ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%s.%s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility",
            module_name, class_name);
        #if PY_VERSION_HEX < 0x02050000
        if (PyErr_Warn(NULL, warning) < 0) goto bad;
        #else
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
        #endif
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%s.%s has the wrong size, try recompiling",
            module_name, class_name);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = (start + end) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,            /*int argcount,*/
        0,            /*int kwonlyargcount,*/
        0,            /*int nlocals,*/
        0,            /*int stacksize,*/
        0,            /*int flags,*/
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,      /*int firstlineno,*/
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_globals = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_globals = PyModule_GetDict(__pyx_m);
    if (!py_globals) goto bad;
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        py_globals,          /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else  /* Python 3+ has unicode identifiers */
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return Py_INCREF(x), x;
  m = Py_TYPE(x)->tp_as_number;
#if PY_VERSION_HEX < 0x03000000
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_VERSION_HEX < 0x03000000
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%s__ returned non-%s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject* x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
#if PY_VERSION_HEX < 0x02050000
   if (ival <= LONG_MAX)
       return PyInt_FromLong((long)ival);
   else {
       unsigned char *bytes = (unsigned char *) &ival;
       int one = 1; int little = (int)*(unsigned char*)&one;
       return _PyLong_FromByteArray(bytes, sizeof(size_t), little, 0);
   }
#else
   return PyInt_FromSize_t(ival);
#endif
}
static CYTHON_INLINE size_t __Pyx_PyInt_AsSize_t(PyObject* x) {
   unsigned PY_LONG_LONG val = __Pyx_PyInt_AsUnsignedLongLong(x);
   if (unlikely(val == (unsigned PY_LONG_LONG)-1 && PyErr_Occurred())) {
       return (size_t)-1;
   } else if (unlikely(val != (unsigned PY_LONG_LONG)(size_t)val)) {
       PyErr_SetString(PyExc_OverflowError,
                       "value too large to convert to size_t");
       return (size_t)-1;
   }
   return (size_t)val;
}


#endif /* Py_PYTHON_H */
